# This file is part of cloud-init. See LICENSE file for license information.
# pylint: disable=attribute-defined-outside-init

import copy
import errno
import ipaddress
import logging
import os
from pathlib import Path
from typing import Optional
from unittest import mock

import pytest

import cloudinit.net as net
from cloudinit import subp
from cloudinit.net.ephemeral import EphemeralIPv4Network, EphemeralIPv6Network
from cloudinit.subp import ProcessExecutionError
from cloudinit.util import ensure_file, write_file
from tests.unittests.helpers import (
    assert_count_equal,
    example_netdev,
    random_string,
)
from tests.unittests.util import MockDistro


class TestSysDevPath:
    def test_sys_dev_path(self):
        """sys_dev_path returns a path under SYS_CLASS_NET for a device."""
        dev = "something"
        path = "attribute"
        expected = net.get_sys_class_path() + dev + "/" + path
        assert expected == net.sys_dev_path(dev, path)

    def test_sys_dev_path_without_path(self):
        """When path param isn't provided it defaults to empty string."""
        dev = "something"
        expected = net.get_sys_class_path() + dev + "/"
        assert expected == net.sys_dev_path(dev)


class TestReadSysNet:
    @pytest.fixture(autouse=True)
    def setup(self, tmpdir_factory):
        # We mock individual numbered tmpdirs here because these tests write
        # to the sysfs directory and stale test artifacts break later tests.
        mock_sysfs = f"{tmpdir_factory.mktemp('sysfs', numbered=True)}/"
        with mock.patch(
            "cloudinit.net.get_sys_class_path", return_value=mock_sysfs
        ):
            self.sysdir = mock_sysfs
            yield

    def test_read_sys_net_strips_contents_of_sys_path(self):
        """read_sys_net strips whitespace from the contents of a sys file."""
        content = "some stuff with trailing whitespace\t\r\n"
        write_file(os.path.join(self.sysdir, "dev", "attr"), content)
        assert content.strip() == net.read_sys_net("dev", "attr")

    def test_read_sys_net_reraises_oserror(self):
        """read_sys_net raises OSError/IOError when file doesn't exist."""
        # Non-specific Exception because versions of python OSError vs IOError.
        with pytest.raises(Exception, match="No such file or directory"):
            net.read_sys_net("dev", "attr")

    def test_read_sys_net_handles_error_with_on_enoent(self):
        """read_sys_net handles OSError/IOError with on_enoent if provided."""
        handled_errors = []

        def on_enoent(e):
            handled_errors.append(e)

        net.read_sys_net("dev", "attr", on_enoent=on_enoent)
        error = handled_errors[0]
        assert isinstance(error, Exception)
        assert "No such file or directory" in str(error)

    def test_read_sys_net_translates_content(self):
        """read_sys_net translates content when translate dict is provided."""
        content = "you're welcome\n"
        write_file(os.path.join(self.sysdir, "dev", "attr"), content)
        translate = {"you're welcome": "de nada"}
        assert "de nada" == net.read_sys_net(
            "dev", "attr", translate=translate
        )

    def test_read_sys_net_errors_on_translation_failures(self, caplog):
        """read_sys_net raises a KeyError and logs details on failure."""
        content = "you're welcome\n"
        write_file(os.path.join(self.sysdir, "dev", "attr"), content)
        with pytest.raises(KeyError, match='"you\'re welcome"'):
            net.read_sys_net("dev", "attr", translate={})
        assert (
            "Found unexpected (not translatable) value 'you're welcome' in "
            "'{0}dev/attr".format(self.sysdir) in caplog.text
        )

    def test_read_sys_net_handles_handles_with_onkeyerror(self):
        """read_sys_net handles translation errors calling on_keyerror."""
        content = "you're welcome\n"
        write_file(os.path.join(self.sysdir, "dev", "attr"), content)
        handled_errors = []

        def on_keyerror(e):
            handled_errors.append(e)

        net.read_sys_net("dev", "attr", translate={}, on_keyerror=on_keyerror)
        error = handled_errors[0]
        assert isinstance(error, KeyError)
        assert '"you\'re welcome"' == str(error)

    def test_read_sys_net_safe_false_on_translate_failure(self):
        """read_sys_net_safe returns False on translation failures."""
        content = "you're welcome\n"
        write_file(os.path.join(self.sysdir, "dev", "attr"), content)
        assert not net.read_sys_net_safe("dev", "attr", translate={})

    def test_read_sys_net_safe_returns_false_on_noent_failure(self):
        """read_sys_net_safe returns False on file not found failures."""
        assert not net.read_sys_net_safe("dev", "attr")

    def test_read_sys_net_int_returns_none_on_error(self):
        """read_sys_net_safe returns None on failures."""
        assert not net.read_sys_net_int("dev", "attr")

    def test_read_sys_net_int_returns_none_on_valueerror(self):
        """read_sys_net_safe returns None when content is not an int."""
        write_file(os.path.join(self.sysdir, "dev", "attr"), "NOTINT\n")
        assert not net.read_sys_net_int("dev", "attr")

    def test_read_sys_net_int_returns_integer_from_content(self):
        """read_sys_net_safe returns None on failures."""
        write_file(os.path.join(self.sysdir, "dev", "attr"), "1\n")
        assert 1 == net.read_sys_net_int("dev", "attr")

    def test_is_up_true(self):
        """is_up is True if sys/net/devname/operstate is 'up' or 'unknown'."""
        for state in ["up", "unknown"]:
            write_file(os.path.join(self.sysdir, "eth0", "operstate"), state)
            assert net.is_up("eth0")

    def test_is_up_false(self):
        """is_up is False if sys/net/devname/operstate is 'down' or invalid."""
        for state in ["down", "incomprehensible"]:
            write_file(os.path.join(self.sysdir, "eth0", "operstate"), state)
            assert not net.is_up("eth0")

    def test_is_bridge(self):
        """is_bridge is True when /sys/net/devname/bridge exists."""
        assert not net.is_bridge("eth0")
        ensure_file(os.path.join(self.sysdir, "eth0", "bridge"))
        assert net.is_bridge("eth0")

    def test_is_bond(self):
        """is_bond is True when /sys/net/devname/bonding exists."""
        assert not net.is_bond("eth0")
        ensure_file(os.path.join(self.sysdir, "eth0", "bonding"))
        assert net.is_bond("eth0")

    def test_get_master(self):
        """get_master returns the path when /sys/net/devname/master exists."""
        assert net.get_master("enP1s1") is None
        master_path = os.path.join(self.sysdir, "enP1s1", "master")
        ensure_file(master_path)
        assert master_path == net.get_master("enP1s1")

    def test_master_is_bridge_or_bond(self):
        bridge_mac = "aa:bb:cc:aa:bb:cc"
        bond_mac = "cc:bb:aa:cc:bb:aa"

        # No master => False
        write_file(os.path.join(self.sysdir, "eth1", "address"), bridge_mac)
        write_file(os.path.join(self.sysdir, "eth2", "address"), bond_mac)

        assert not net.master_is_bridge_or_bond("eth1")
        assert not net.master_is_bridge_or_bond("eth2")

        # masters without bridge/bonding => False
        write_file(os.path.join(self.sysdir, "br0", "address"), bridge_mac)
        write_file(os.path.join(self.sysdir, "bond0", "address"), bond_mac)

        os.symlink("../br0", os.path.join(self.sysdir, "eth1", "master"))
        os.symlink("../bond0", os.path.join(self.sysdir, "eth2", "master"))

        assert not net.master_is_bridge_or_bond("eth1")
        assert not net.master_is_bridge_or_bond("eth2")

        # masters with bridge/bonding => True
        write_file(os.path.join(self.sysdir, "br0", "bridge"), "")
        write_file(os.path.join(self.sysdir, "bond0", "bonding"), "")

        assert net.master_is_bridge_or_bond("eth1")
        assert net.master_is_bridge_or_bond("eth2")

    def test_master_is_openvswitch(self):
        ovs_mac = "bb:cc:aa:bb:cc:aa"

        # No master => False
        write_file(os.path.join(self.sysdir, "eth1", "address"), ovs_mac)

        assert not net.master_is_bridge_or_bond("eth1")

        # masters without ovs-system => False
        write_file(os.path.join(self.sysdir, "ovs-system", "address"), ovs_mac)

        os.symlink(
            "../ovs-system", os.path.join(self.sysdir, "eth1", "master")
        )

        assert not net.master_is_openvswitch("eth1")

        # masters with ovs-system => True
        os.symlink(
            "../ovs-system",
            os.path.join(self.sysdir, "eth1", "upper_ovs-system"),
        )

        assert net.master_is_openvswitch("eth1")

    def test_is_vlan(self):
        """is_vlan is True when /sys/net/devname/uevent has DEVTYPE=vlan."""
        ensure_file(os.path.join(self.sysdir, "eth0", "uevent"))
        assert not net.is_vlan("eth0")
        content = "junk\nDEVTYPE=vlan\njunk\n"
        write_file(os.path.join(self.sysdir, "eth0", "uevent"), content)
        assert net.is_vlan("eth0")


class TestGenerateFallbackConfig:
    @pytest.fixture(autouse=True)
    def fixtures(self, mocker, tmp_path):
        self.sysdir = str(tmp_path) + "/"
        mocker.patch(
            "cloudinit.net.get_sys_class_path", return_value=self.sysdir
        )
        mocker.patch("cloudinit.net.util.is_container", return_value=False)
        mocker.patch("cloudinit.net.util.udevadm_settle")
        self.m_is_netfail = mocker.patch(
            "cloudinit.net.is_netfailover", return_value=False
        )
        mocker.patch(
            "cloudinit.net.is_netfail_master",
            return_value=False,
        )

    def test_generate_fallback_finds_connected_eth_with_mac(self):
        """generate_fallback_config finds any connected device with a mac."""
        write_file(os.path.join(self.sysdir, "eth0", "carrier"), "1")
        write_file(os.path.join(self.sysdir, "eth1", "carrier"), "1")
        mac = "aa:bb:cc:aa:bb:cc"
        write_file(os.path.join(self.sysdir, "eth1", "address"), mac)
        expected = {
            "ethernets": {
                "eth1": {
                    "match": {"macaddress": mac},
                    "dhcp4": True,
                    "dhcp6": True,
                    "set-name": "eth1",
                }
            },
            "version": 2,
        }
        assert expected == net.generate_fallback_config()

    def test_generate_fallback_finds_dormant_eth_with_mac(self):
        """generate_fallback_config finds any dormant device with a mac."""
        write_file(os.path.join(self.sysdir, "eth0", "dormant"), "1")
        mac = "aa:bb:cc:aa:bb:cc"
        write_file(os.path.join(self.sysdir, "eth0", "address"), mac)
        expected = {
            "ethernets": {
                "eth0": {
                    "match": {"macaddress": mac},
                    "dhcp4": True,
                    "dhcp6": True,
                    "set-name": "eth0",
                }
            },
            "version": 2,
        }
        assert expected == net.generate_fallback_config()

    def test_generate_fallback_finds_eth_by_operstate(self):
        """generate_fallback_config finds any dormant device with a mac."""
        mac = "aa:bb:cc:aa:bb:cc"
        write_file(os.path.join(self.sysdir, "eth0", "address"), mac)
        expected = {
            "ethernets": {
                "eth0": {
                    "dhcp4": True,
                    "dhcp6": True,
                    "match": {"macaddress": mac},
                    "set-name": "eth0",
                }
            },
            "version": 2,
        }
        valid_operstates = ["dormant", "down", "lowerlayerdown", "unknown"]
        for state in valid_operstates:
            write_file(os.path.join(self.sysdir, "eth0", "operstate"), state)
            assert expected == net.generate_fallback_config()
        write_file(os.path.join(self.sysdir, "eth0", "operstate"), "noworky")
        assert net.generate_fallback_config() is None

    def test_generate_fallback_config_skips_veth(self):
        """generate_fallback_config will skip any veth interfaces."""
        # A connected veth which gets ignored
        write_file(os.path.join(self.sysdir, "veth0", "carrier"), "1")
        assert net.generate_fallback_config() is None

    def test_generate_fallback_config_skips_bridges(self):
        """generate_fallback_config will skip any bridges interfaces."""
        # A connected veth which gets ignored
        write_file(os.path.join(self.sysdir, "eth0", "carrier"), "1")
        mac = "aa:bb:cc:aa:bb:cc"
        write_file(os.path.join(self.sysdir, "eth0", "address"), mac)
        ensure_file(os.path.join(self.sysdir, "eth0", "bridge"))
        assert net.generate_fallback_config() is None

    def test_generate_fallback_config_skips_bonds(self):
        """generate_fallback_config will skip any bonded interfaces."""
        # A connected veth which gets ignored
        write_file(os.path.join(self.sysdir, "eth0", "carrier"), "1")
        mac = "aa:bb:cc:aa:bb:cc"
        write_file(os.path.join(self.sysdir, "eth0", "address"), mac)
        ensure_file(os.path.join(self.sysdir, "eth0", "bonding"))
        assert net.generate_fallback_config() is None

    @mock.patch("cloudinit.net.is_netfail_master")
    def test_generate_fallback_config_skips_netfail_devs(
        self, m_is_netfail_master
    ):
        """gen_fallback_config ignores netfail primary,sby no mac on master."""
        mac = "aa:bb:cc:aa:bb:cc"  # netfailover devs share the same mac
        for iface in ["ens3", "ens3sby", "enP0s1f3"]:
            write_file(os.path.join(self.sysdir, iface, "carrier"), "1")
            write_file(
                os.path.join(self.sysdir, iface, "addr_assign_type"), "0"
            )
            write_file(os.path.join(self.sysdir, iface, "address"), mac)

        def is_netfail(iface, _driver=None):
            # ens3 is the master
            if iface == "ens3":
                return False
            return True

        self.m_is_netfail.side_effect = is_netfail

        def is_netfail_master(iface, _driver=None):
            # ens3 is the master
            if iface == "ens3":
                return True
            return False

        m_is_netfail_master.side_effect = is_netfail_master
        expected = {
            "ethernets": {
                "ens3": {
                    "dhcp4": True,
                    "dhcp6": True,
                    "match": {"name": "ens3"},
                    "set-name": "ens3",
                }
            },
            "version": 2,
        }
        result = net.generate_fallback_config()
        assert expected == result


class TestNetFindFallBackNic:
    @pytest.fixture(autouse=True)
    def fixtures(self, mocker, tmp_path):
        self.sysdir = str(tmp_path) + "/"
        mocker.patch(
            "cloudinit.net.get_sys_class_path", return_value=self.sysdir
        )
        mocker.patch(
            "cloudinit.net.util.is_container",
            return_value=False,
        )
        mocker.patch("cloudinit.net.util.udevadm_settle")

    def test_generate_fallback_finds_first_connected_eth_with_mac(self):
        """find_fallback_nic finds any connected device with a mac."""
        write_file(os.path.join(self.sysdir, "eth0", "carrier"), "1")
        write_file(os.path.join(self.sysdir, "eth1", "carrier"), "1")
        mac = "aa:bb:cc:aa:bb:cc"
        write_file(os.path.join(self.sysdir, "eth1", "address"), mac)
        assert "eth1" == net.find_fallback_nic()


class TestNetFindCandidateNics:
    def create_fake_interface(
        self,
        name: str,
        address: Optional[str] = "aa:bb:cc:aa:bb:cc",
        carrier: bool = True,
        bonding: bool = False,
        dormant: bool = False,
        driver: str = "fakenic",
        bridge: bool = False,
        failover_standby: bool = False,
        operstate: Optional[str] = None,
        renamed: bool = True,
    ):
        interface_path = self.sys_path / name
        interface_path.mkdir(parents=True)

        if address is not None:
            (interface_path / "address").write_text(str(address))

        if carrier:
            (interface_path / "carrier").write_text("1")
        else:
            (interface_path / "carrier").write_text("0")

        if bonding:
            (interface_path / "bonding").write_text("1")

        if bridge:
            (interface_path / "bridge").write_text("1")

        if dormant:
            (interface_path / "dormant").write_text("1")
        else:
            (interface_path / "dormant").write_text("0")

        if operstate:
            (interface_path / "operstate").write_text(operstate)

        device_path = interface_path / "device"
        device_path.mkdir()
        if failover_standby:
            driver = "virtio_net"
            (interface_path / "master").symlink_to(os.path.join("..", name))
            (device_path / "features").write_text("1" * 64)

        if driver:
            (device_path / driver).write_text(driver)
            (device_path / "driver").symlink_to(driver)

        if renamed:
            (interface_path / "name_assign_type").write_text("4")
        else:
            (interface_path / "name_assign_type").write_text("0")

    @pytest.fixture(autouse=True)
    def setup(self, monkeypatch, tmpdir):
        """Fixture to set up the test environment."""
        self.sys_path = Path(tmpdir) / "sys"
        monkeypatch.setattr(
            net, "get_sys_class_path", lambda: str(self.sys_path) + "/"
        )
        monkeypatch.setattr(
            net.util,
            "is_container",
            lambda: False,
        )

        self.m_settle = mock.Mock(return_value=None)
        self.m_cmdline = mock.Mock(return_value="")
        monkeypatch.setattr("cloudinit.net.util.udevadm_settle", self.m_settle)
        monkeypatch.setattr("cloudinit.net.util.get_cmdline", self.m_cmdline)

    def test_ignored_interfaces(self):
        self.create_fake_interface(
            name="ethNoCarrierDormantOperstateIgnored",
            carrier=False,
        )
        self.create_fake_interface(
            name="ethWithoutMacIgnored",
            address=None,
        )
        self.create_fake_interface(name="vethIgnored", carrier=1)
        self.create_fake_interface(
            name="bondIgnored",
            bonding=True,
        )
        self.create_fake_interface(
            name="bridgeIgnored",
            bridge=True,
        )
        self.create_fake_interface(
            name="failOverIgnored",
            failover_standby=True,
        )
        self.create_fake_interface(
            name="TestingOperStateIgnored",
            carrier=False,
            operstate="testing",
        )
        self.create_fake_interface(
            name="hv",
            driver="hv_netvsc",
            address="00:11:22:00:00:f0",
        )
        self.create_fake_interface(
            name="hv_vf_mlx4",
            driver="mlx4_core",
            address="00:11:22:00:00:f0",
        )
        self.create_fake_interface(
            name="hv_vf_mlx5",
            driver="mlx5_core",
            address="00:11:22:00:00:f0",
        )
        self.create_fake_interface(
            name="hv_vf_mana",
            driver="mana",
            address="00:11:22:00:00:f0",
        )

        assert net.find_candidate_nics_on_linux() == ["hv"]

    def test_carrier_preferred(self):
        self.create_fake_interface(name="eth0", carrier=False, dormant=True)
        self.create_fake_interface(name="eth1")

        assert net.find_candidate_nics_on_linux() == ["eth1", "eth0"]

    def test_natural_sort(self):
        self.create_fake_interface(name="a")
        self.create_fake_interface(name="a1")
        self.create_fake_interface(name="a2")
        self.create_fake_interface(name="a10")
        self.create_fake_interface(name="b1")

        assert net.find_candidate_nics_on_linux() == [
            "a",
            "a1",
            "a2",
            "a10",
            "b1",
        ]

    def test_eth0_preferred_with_carrier(self):
        self.create_fake_interface(name="abc0")
        self.create_fake_interface(name="eth0")

        assert net.find_candidate_nics_on_linux() == ["eth0", "abc0"]

    @pytest.mark.parametrize("dormant", [False, True])
    @pytest.mark.parametrize(
        "operstate", ["dormant", "down", "lowerlayerdown", "unknown"]
    )
    def test_eth0_preferred_after_carrier(self, dormant, operstate):
        self.create_fake_interface(name="xeth10")
        self.create_fake_interface(name="eth", carrier=False, dormant=True)
        self.create_fake_interface(
            name="eth0",
            carrier=False,
            dormant=dormant,
            operstate=operstate,
        )
        self.create_fake_interface(name="eth1", carrier=False, dormant=True)
        self.create_fake_interface(
            name="eth2",
            carrier=False,
            operstate=operstate,
        )

        assert net.find_candidate_nics_on_linux() == [
            "xeth10",
            "eth0",
            "eth",
            "eth1",
            "eth2",
        ]

    def test_no_nics(self):
        assert net.find_candidate_nics_on_linux() == []

    def test_udevadm_settle_called_for_predictable_names(self):
        self.create_fake_interface(name="eth0", renamed=False)

        result = net.find_candidate_nics_on_linux()

        assert result == ["eth0"]
        self.m_settle.assert_called_once()

    def test_udevadm_settle_not_called_for_unpredictable_names(self):
        self.create_fake_interface(name="eth0", renamed=False)
        self.m_cmdline.return_value = "net.ifnames=0 biosdevname=0"

        result = net.find_candidate_nics_on_linux()

        assert result == ["eth0"]
        self.m_settle.assert_not_called()

    def test_udevadm_settle_not_called_for_renamed_interface(self):
        self.create_fake_interface(name="eth0", renamed=True)

        result = net.find_candidate_nics_on_linux()

        assert result == ["eth0"]
        self.m_settle.assert_not_called()

    def test_udevadm_settle_failure_handled_gracefully(self, caplog):
        self.create_fake_interface(name="eth0", renamed=False)

        self.m_settle.side_effect = subp.ProcessExecutionError(
            cmd="xcmd", stderr="xstderr", stdout="xstdout", exit_code=1
        )

        with caplog.at_level(logging.WARNING):
            result = net.find_candidate_nics_on_linux()
            assert (
                "udevadm failed to settle: cmd='xcmd' "
                "stderr='xstderr' stdout='xstdout' exit_code=1" in caplog.text
            )

        assert result == ["eth0"]
        self.m_settle.assert_called_once()


class TestGetDeviceList:
    @pytest.fixture(autouse=True)
    def fixtures(self, mocker, tmp_path):
        self.sysdir = str(tmp_path) + "/"
        self.m_sys_path = mocker.patch(
            "cloudinit.net.get_sys_class_path", return_value=self.sysdir
        )

    def test_get_devicelist_raise_oserror(self):
        """get_devicelist raise any non-ENOENT OSerror."""
        error = OSError("Can not do it")
        error.errno = errno.EPERM  # Set non-ENOENT
        self.m_sys_path.side_effect = error
        with pytest.raises(OSError, match="Can not do it"):
            net.get_devicelist()

    def test_get_devicelist_empty_without_sys_net(self):
        """get_devicelist returns empty list when missing SYS_CLASS_NET."""
        self.m_sys_path.return_value = "idontexist"
        assert [] == net.get_devicelist()

    def test_get_devicelist_empty_with_no_devices_in_sys_net(self):
        """get_devicelist returns empty directory listing for SYS_CLASS_NET."""
        assert [] == net.get_devicelist()

    def test_get_devicelist_lists_any_subdirectories_in_sys_net(self):
        """get_devicelist returns a directory listing for SYS_CLASS_NET."""
        write_file(os.path.join(self.sysdir, "eth0", "operstate"), "up")
        write_file(os.path.join(self.sysdir, "eth1", "operstate"), "up")
        assert_count_equal(["eth0", "eth1"], net.get_devicelist())


@mock.patch(
    "cloudinit.net.is_openvswitch_internal_interface",
    mock.Mock(return_value=False),
)
class TestGetInterfaceMAC:
    @pytest.fixture(autouse=True)
    def fixtures(self, mocker, tmp_path):
        self.sysdir = str(tmp_path) + "/"
        self.m_sys_path = mocker.patch(
            "cloudinit.net.get_sys_class_path", return_value=self.sysdir
        )

    def test_get_interface_mac_false_with_no_mac(self):
        """get_device_list returns False when no mac is reported."""
        ensure_file(os.path.join(self.sysdir, "eth0", "bonding"))
        mac_path = os.path.join(self.sysdir, "eth0", "address")
        assert not os.path.exists(mac_path)
        assert net.get_interface_mac("eth0") is False

    def test_get_interface_mac(self):
        """get_interfaces returns the mac from SYS_CLASS_NET/dev/address."""
        mac = "aa:bb:cc:aa:bb:cc"
        write_file(os.path.join(self.sysdir, "eth1", "address"), mac)
        assert mac == net.get_interface_mac("eth1")

    def test_get_interface_mac_grabs_bonding_address(self):
        """get_interfaces returns the source device mac for bonded devices."""
        source_dev_mac = "aa:bb:cc:aa:bb:cc"
        bonded_mac = "dd:ee:ff:dd:ee:ff"
        write_file(os.path.join(self.sysdir, "eth1", "address"), bonded_mac)
        write_file(
            os.path.join(self.sysdir, "eth1", "bonding_slave", "perm_hwaddr"),
            source_dev_mac,
        )
        assert source_dev_mac == net.get_interface_mac("eth1")

    def test_get_interfaces_empty_list_without_sys_net(self):
        """get_interfaces returns an empty list when missing SYS_CLASS_NET."""
        self.m_sys_path.return_value = "idontexist"
        assert [] == net.get_interfaces()

    def test_get_interfaces_by_mac_skips_empty_mac(self):
        """Ignore 00:00:00:00:00:00 addresses from get_interfaces_by_mac."""
        empty_mac = "00:00:00:00:00:00"
        mac = "aa:bb:cc:aa:bb:cc"
        write_file(os.path.join(self.sysdir, "eth1", "address"), empty_mac)
        write_file(os.path.join(self.sysdir, "eth1", "addr_assign_type"), "0")
        write_file(os.path.join(self.sysdir, "eth2", "addr_assign_type"), "0")
        write_file(os.path.join(self.sysdir, "eth2", "address"), mac)
        expected = [("eth2", "aa:bb:cc:aa:bb:cc", None, None)]
        assert expected == net.get_interfaces()

    def test_get_interfaces_by_mac_skips_missing_mac(self):
        """Ignore interfaces without an address from get_interfaces_by_mac."""
        write_file(os.path.join(self.sysdir, "eth1", "addr_assign_type"), "0")
        address_path = os.path.join(self.sysdir, "eth1", "address")
        assert not os.path.exists(address_path)
        mac = "aa:bb:cc:aa:bb:cc"
        write_file(os.path.join(self.sysdir, "eth2", "addr_assign_type"), "0")
        write_file(os.path.join(self.sysdir, "eth2", "address"), mac)
        expected = [("eth2", "aa:bb:cc:aa:bb:cc", None, None)]
        assert expected == net.get_interfaces()

    def test_get_interfaces_by_mac_skips_master_devs(self):
        """Ignore interfaces with a master device which would have dup mac."""
        mac1 = mac2 = "aa:bb:cc:aa:bb:cc"
        write_file(os.path.join(self.sysdir, "eth1", "addr_assign_type"), "0")
        write_file(os.path.join(self.sysdir, "eth1", "address"), mac1)
        write_file(os.path.join(self.sysdir, "eth1", "master"), "blah")
        write_file(os.path.join(self.sysdir, "eth2", "addr_assign_type"), "0")
        write_file(os.path.join(self.sysdir, "eth2", "address"), mac2)
        expected = [("eth2", mac2, None, None)]
        assert expected == net.get_interfaces()

    @mock.patch("cloudinit.net.is_netfailover")
    def test_get_interfaces_by_mac_skips_netfailvoer(self, m_netfail):
        """Ignore interfaces if netfailover primary or standby."""
        mac = "aa:bb:cc:aa:bb:cc"  # netfailover devs share the same mac
        for iface in ["ens3", "ens3sby", "enP0s1f3"]:
            write_file(
                os.path.join(self.sysdir, iface, "addr_assign_type"), "0"
            )
            write_file(os.path.join(self.sysdir, iface, "address"), mac)

        def is_netfail(iface, _driver=None):
            # ens3 is the master
            if iface == "ens3":
                return False
            else:
                return True

        m_netfail.side_effect = is_netfail
        expected = [("ens3", mac, None, None)]
        assert expected == net.get_interfaces()

    def test_get_interfaces_does_not_skip_phys_members_of_bridges_and_bonds(
        self,
    ):
        bridge_mac = "aa:bb:cc:aa:bb:cc"
        bond_mac = "cc:bb:aa:cc:bb:aa"
        ovs_mac = "bb:cc:aa:bb:cc:aa"

        write_file(os.path.join(self.sysdir, "br0", "address"), bridge_mac)
        write_file(os.path.join(self.sysdir, "br0", "bridge"), "")

        write_file(os.path.join(self.sysdir, "bond0", "address"), bond_mac)
        write_file(os.path.join(self.sysdir, "bond0", "bonding"), "")

        write_file(os.path.join(self.sysdir, "ovs-system", "address"), ovs_mac)

        write_file(os.path.join(self.sysdir, "eth1", "address"), bridge_mac)
        os.symlink("../br0", os.path.join(self.sysdir, "eth1", "master"))

        write_file(os.path.join(self.sysdir, "eth2", "address"), bond_mac)
        os.symlink("../bond0", os.path.join(self.sysdir, "eth2", "master"))

        write_file(os.path.join(self.sysdir, "eth3", "address"), ovs_mac)
        os.symlink(
            "../ovs-system", os.path.join(self.sysdir, "eth3", "master")
        )
        os.symlink(
            "../ovs-system",
            os.path.join(self.sysdir, "eth3", "upper_ovs-system"),
        )

        interface_names = [interface[0] for interface in net.get_interfaces()]
        assert ["eth1", "eth2", "eth3", "ovs-system"] == sorted(
            interface_names
        )


class TestInterfaceHasOwnMAC:
    @pytest.fixture(autouse=True)
    def fixtures(self, mocker, tmp_path):
        self.sysdir = str(tmp_path) + "/"
        mocker.patch(
            "cloudinit.net.get_sys_class_path", return_value=self.sysdir
        )

    def test_interface_has_own_mac_false_when_stolen(self):
        """Return False from interface_has_own_mac when address is stolen."""
        write_file(os.path.join(self.sysdir, "eth1", "addr_assign_type"), "2")
        assert net.interface_has_own_mac("eth1") is False

    def test_interface_has_own_mac_true_when_not_stolen(self):
        """Return False from interface_has_own_mac when mac isn't stolen."""
        valid_assign_types = ["0", "1", "3"]
        assign_path = os.path.join(self.sysdir, "eth1", "addr_assign_type")
        for _type in valid_assign_types:
            write_file(assign_path, _type)
            assert net.interface_has_own_mac("eth1") is True

    def test_interface_has_own_mac_strict_errors_on_absent_assign_type(self):
        """When addr_assign_type is absent, interface_has_own_mac errors."""
        with pytest.raises(ValueError):
            net.interface_has_own_mac("eth1", strict=True)


@mock.patch("cloudinit.net.subp.subp")
@pytest.mark.usefixtures("disable_netdev_info")
class TestEphemeralIPV4Network:
    @pytest.fixture(autouse=True)
    def fixtures(self, mocker, tmp_path):
        self.sysdir = str(tmp_path) + "/"
        mocker.patch(
            "cloudinit.net.get_sys_class_path", return_value=self.sysdir
        )

    def test_ephemeral_ipv4_network_errors_on_missing_params(self, m_subp):
        """No required params for EphemeralIPv4Network can be None."""
        required_params = {
            "interface": "eth0",
            "ip": "192.168.2.2",
            "prefix_or_mask": "255.255.255.0",
            "broadcast": "192.168.2.255",
        }
        for key in required_params.keys():
            params = copy.deepcopy(required_params)
            params[key] = None
            with pytest.raises(ValueError, match="Cannot init network on"):
                EphemeralIPv4Network(
                    MockDistro(),
                    interface_addrs_before_dhcp=example_netdev,
                    **params,
                )
            assert 0 == m_subp.call_count

    def test_ephemeral_ipv4_network_errors_invalid_mask_prefix(self, m_subp):
        """Raise an error when prefix_or_mask is not a netmask or prefix."""
        params = {
            "interface": "eth0",
            "ip": "192.168.2.2",
            "broadcast": "192.168.2.255",
            "interface_addrs_before_dhcp": example_netdev,
        }
        invalid_masks = ("invalid", "invalid.", "123.123.123")
        for error_val in invalid_masks:
            params["prefix_or_mask"] = error_val
            with pytest.raises(
                ValueError,
                match="Cannot setup network, invalid prefix or netmask: ",
            ):
                with EphemeralIPv4Network(MockDistro(), **params):
                    pass
            assert 0 == m_subp.call_count

    def test_ephemeral_ipv4_network_performs_teardown(self, m_subp):
        """EphemeralIPv4Network performs teardown on the device if setup."""
        expected_setup_calls = [
            mock.call(
                [
                    "ip",
                    "-family",
                    "inet",
                    "addr",
                    "add",
                    "192.168.2.2/24",
                    "broadcast",
                    "192.168.2.255",
                    "dev",
                    "eth0",
                ],
            ),
        ]
        expected_teardown_calls = [
            mock.call(
                [
                    "ip",
                    "-family",
                    "inet",
                    "addr",
                    "del",
                    "192.168.2.2/24",
                    "dev",
                    "eth0",
                ],
            ),
        ]
        params = {
            "interface": "eth0",
            "ip": "192.168.2.2",
            "prefix_or_mask": "255.255.255.0",
            "broadcast": "192.168.2.255",
            "interface_addrs_before_dhcp": example_netdev,
        }
        with EphemeralIPv4Network(MockDistro(), **params):
            assert expected_setup_calls == m_subp.call_args_list
        m_subp.assert_has_calls(expected_teardown_calls)

    def test_teardown_on_enter_exception(self, m_subp):
        """Ensure ephemeral teardown happens.

        Even though we're using a context manager, we need to handle any
        exceptions raised in __enter__ manually and do the appropriate
        teardown.
        """

        def side_effect(args, **kwargs):
            if "append" in args and "3.3.3.3/32" in args:
                raise subp.ProcessExecutionError("oh no!")

        m_subp.side_effect = side_effect

        with pytest.raises(subp.ProcessExecutionError):
            with EphemeralIPv4Network(
                MockDistro(),
                interface="eth0",
                ip="1.1.1.1",
                prefix_or_mask="255.255.255.0",
                broadcast="1.1.1.255",
                interface_addrs_before_dhcp=example_netdev,
                static_routes=[
                    ("2.2.2.2/32", "9.9.9.9"),
                    ("3.3.3.3/32", "8.8.8.8"),
                ],
            ):
                pass

        expected_teardown_calls = [
            mock.call(
                [
                    "ip",
                    "-4",
                    "route",
                    "del",
                    "2.2.2.2/32",
                    "via",
                    "9.9.9.9",
                    "dev",
                    "eth0",
                ],
            ),
            mock.call(
                [
                    "ip",
                    "-family",
                    "inet",
                    "addr",
                    "del",
                    "1.1.1.1/24",
                    "dev",
                    "eth0",
                ],
            ),
        ]
        for teardown in expected_teardown_calls:
            assert teardown in m_subp.call_args_list

    def test_ephemeral_ipv4_network_noop_when_configured(self, m_subp, caplog):
        """EphemeralIPv4Network handles exception when address is setup.

        It performs no cleanup as the interface was already setup.
        """
        params = {
            "interface": "eth0",
            "ip": "10.85.130.116",
            "prefix_or_mask": "255.255.255.0",
            "broadcast": "192.168.2.255",
            "interface_addrs_before_dhcp": example_netdev,
        }
        m_subp.side_effect = ProcessExecutionError(
            "", "RTNETLINK answers: File exists", 2
        )
        expected_calls = []
        with EphemeralIPv4Network(MockDistro(), **params):
            pass
        assert expected_calls == m_subp.call_args_list
        assert "Skip bringing up network link" in caplog.text
        assert "Skip adding ip address" in caplog.text

    def test_ephemeral_ipv4_network_with_prefix(self, m_subp):
        """EphemeralIPv4Network takes a valid prefix to setup the network."""
        params = {
            "interface": "eth0",
            "ip": "192.168.2.2",
            "prefix_or_mask": "24",
            "broadcast": "192.168.2.255",
            "interface_addrs_before_dhcp": example_netdev,
        }
        for prefix_val in ["24", 16]:  # prefix can be int or string
            params["prefix_or_mask"] = prefix_val
            with EphemeralIPv4Network(MockDistro(), **params):
                pass
        m_subp.assert_has_calls(
            [
                mock.call(
                    [
                        "ip",
                        "-family",
                        "inet",
                        "addr",
                        "add",
                        "192.168.2.2/24",
                        "broadcast",
                        "192.168.2.255",
                        "dev",
                        "eth0",
                    ],
                )
            ]
        )
        m_subp.assert_has_calls(
            [
                mock.call(
                    [
                        "ip",
                        "-family",
                        "inet",
                        "addr",
                        "add",
                        "192.168.2.2/16",
                        "broadcast",
                        "192.168.2.255",
                        "dev",
                        "eth0",
                    ],
                )
            ]
        )

    def test_ephemeral_ipv4_network_with_new_default_route(self, m_subp):
        """Add the route when router is set and no default route exists."""
        params = {
            "interface": "eth0",
            "ip": "192.168.2.2",
            "prefix_or_mask": "255.255.255.0",
            "broadcast": "192.168.2.255",
            "router": "192.168.2.1",
            "interface_addrs_before_dhcp": example_netdev,
        }
        # Empty response from ip route gw check
        m_subp.return_value = subp.SubpResult("", "")
        expected_setup_calls = [
            mock.call(
                [
                    "ip",
                    "-family",
                    "inet",
                    "addr",
                    "add",
                    "192.168.2.2/24",
                    "broadcast",
                    "192.168.2.255",
                    "dev",
                    "eth0",
                ],
            ),
            mock.call(["ip", "route", "show", "0.0.0.0/0"]),
            mock.call(
                [
                    "ip",
                    "-4",
                    "route",
                    "replace",
                    "192.168.2.1",
                    "dev",
                    "eth0",
                    "src",
                    "192.168.2.2",
                ],
            ),
            mock.call(
                [
                    "ip",
                    "-4",
                    "route",
                    "replace",
                    "default",
                    "via",
                    "192.168.2.1",
                    "dev",
                    "eth0",
                ],
            ),
        ]
        expected_teardown_calls = [
            mock.call(
                ["ip", "-4", "route", "del", "default", "dev", "eth0"],
            ),
            mock.call(
                [
                    "ip",
                    "-4",
                    "route",
                    "del",
                    "192.168.2.1",
                    "dev",
                    "eth0",
                    "src",
                    "192.168.2.2",
                ],
            ),
        ]

        with EphemeralIPv4Network(MockDistro(), **params):
            assert expected_setup_calls == m_subp.call_args_list
        m_subp.assert_has_calls(expected_teardown_calls)

    def test_ephemeral_ipv4_network_with_rfc3442_static_routes(self, m_subp):
        params = {
            "interface": "eth0",
            "ip": "192.168.2.2",
            "prefix_or_mask": "255.255.255.255",
            "broadcast": "192.168.2.255",
            "static_routes": [
                ("192.168.2.1/32", "0.0.0.0"),
                ("169.254.169.254/32", "192.168.2.1"),
                ("0.0.0.0/0", "192.168.2.1"),
            ],
            "router": "192.168.2.1",
            "interface_addrs_before_dhcp": example_netdev,
        }
        expected_setup_calls = [
            mock.call(
                [
                    "ip",
                    "-family",
                    "inet",
                    "addr",
                    "add",
                    "192.168.2.2/32",
                    "broadcast",
                    "192.168.2.255",
                    "dev",
                    "eth0",
                ],
            ),
            mock.call(
                [
                    "ip",
                    "-4",
                    "route",
                    "append",
                    "192.168.2.1/32",
                    "dev",
                    "eth0",
                ],
            ),
            mock.call(
                [
                    "ip",
                    "-4",
                    "route",
                    "append",
                    "169.254.169.254/32",
                    "via",
                    "192.168.2.1",
                    "dev",
                    "eth0",
                ],
            ),
            mock.call(
                [
                    "ip",
                    "-4",
                    "route",
                    "append",
                    "0.0.0.0/0",
                    "via",
                    "192.168.2.1",
                    "dev",
                    "eth0",
                ],
            ),
        ]
        expected_teardown_calls = [
            mock.call(
                [
                    "ip",
                    "-4",
                    "route",
                    "del",
                    "0.0.0.0/0",
                    "via",
                    "192.168.2.1",
                    "dev",
                    "eth0",
                ],
            ),
            mock.call(
                [
                    "ip",
                    "-4",
                    "route",
                    "del",
                    "169.254.169.254/32",
                    "via",
                    "192.168.2.1",
                    "dev",
                    "eth0",
                ],
            ),
            mock.call(
                ["ip", "-4", "route", "del", "192.168.2.1/32", "dev", "eth0"],
            ),
            mock.call(
                [
                    "ip",
                    "-family",
                    "inet",
                    "addr",
                    "del",
                    "192.168.2.2/32",
                    "dev",
                    "eth0",
                ],
            ),
        ]
        with EphemeralIPv4Network(MockDistro(), **params):
            assert expected_setup_calls == m_subp.call_args_list
        m_subp.assert_has_calls(expected_setup_calls + expected_teardown_calls)


class TestEphemeralIPV6Network:
    @mock.patch("cloudinit.net.read_sys_net")
    @mock.patch("cloudinit.net.subp.subp")
    def test_ephemeral_ipv6_network_performs_setup(self, m_subp, _):
        """EphemeralIPv4Network performs teardown on the device if setup."""
        expected_setup_calls = [
            mock.call(
                ["ip", "link", "set", "dev", "eth0", "up"],
            ),
        ]
        with EphemeralIPv6Network(MockDistro(), interface="eth0"):
            assert expected_setup_calls == m_subp.call_args_list


def _mk_v1_phys(mac, name, driver, device_id):
    v1_cfg = {"type": "physical", "name": name, "mac_address": mac}
    params = {}
    if driver:
        params.update({"driver": driver})
    if device_id:
        params.update({"device_id": device_id})

    if params:
        v1_cfg.update({"params": params})

    return v1_cfg


def _mk_v2_phys(mac, name, driver=None, device_id=None):
    v2_cfg = {"set-name": name, "match": {"macaddress": mac}}
    if driver:
        v2_cfg["match"].update({"driver": driver})
    if device_id:
        v2_cfg["match"].update({"device_id": device_id})

    return v2_cfg


class TestExtractPhysdevs:
    @pytest.fixture(autouse=True)
    def fixtures(self, mocker):
        self.m_driver = mocker.patch("cloudinit.net.device_driver")
        self.m_devid = mocker.patch(
            "cloudinit.net.device_devid",
        )

    def test_extract_physdevs_looks_up_driver_v1(self):
        driver = "virtio"
        self.m_driver.return_value = driver
        physdevs = [
            ["aa:bb:cc:dd:ee:ff", "eth0", None, "0x1000"],
        ]
        netcfg = {
            "version": 1,
            "config": [_mk_v1_phys(*args) for args in physdevs],
        }
        # insert the driver value for verification
        physdevs[0][2] = driver
        assert sorted(physdevs) == sorted(net.extract_physdevs(netcfg))
        self.m_driver.assert_called_with("eth0")

    def test_extract_physdevs_looks_up_driver_v2(self):
        driver = "virtio"
        self.m_driver.return_value = driver
        physdevs = [
            ["aa:bb:cc:dd:ee:ff", "eth0", None, "0x1000"],
        ]
        netcfg = {
            "version": 2,
            "ethernets": {args[1]: _mk_v2_phys(*args) for args in physdevs},
        }
        # insert the driver value for verification
        physdevs[0][2] = driver
        assert sorted(physdevs) == sorted(net.extract_physdevs(netcfg))
        self.m_driver.assert_called_with("eth0")

    def test_extract_physdevs_looks_up_devid_v1(self):
        devid = "0x1000"
        self.m_devid.return_value = devid
        physdevs = [
            ["aa:bb:cc:dd:ee:ff", "eth0", "virtio", None],
        ]
        netcfg = {
            "version": 1,
            "config": [_mk_v1_phys(*args) for args in physdevs],
        }
        # insert the driver value for verification
        physdevs[0][3] = devid
        assert sorted(physdevs) == sorted(net.extract_physdevs(netcfg))
        self.m_devid.assert_called_with("eth0")

    def test_extract_physdevs_looks_up_devid_v2(self):
        devid = "0x1000"
        self.m_devid.return_value = devid
        physdevs = [
            ["aa:bb:cc:dd:ee:ff", "eth0", "virtio", None],
        ]
        netcfg = {
            "version": 2,
            "ethernets": {args[1]: _mk_v2_phys(*args) for args in physdevs},
        }
        # insert the driver value for verification
        physdevs[0][3] = devid
        assert sorted(physdevs) == sorted(net.extract_physdevs(netcfg))
        self.m_devid.assert_called_with("eth0")

    def test_get_v1_type_physical(self):
        physdevs = [
            ["aa:bb:cc:dd:ee:ff", "eth0", "virtio", "0x1000"],
            ["00:11:22:33:44:55", "ens3", "e1000", "0x1643"],
            ["09:87:65:43:21:10", "ens0p1", "mlx4_core", "0:0:1000"],
        ]
        netcfg = {
            "version": 1,
            "config": [_mk_v1_phys(*args) for args in physdevs],
        }
        assert sorted(physdevs) == sorted(net.extract_physdevs(netcfg))

    def test_get_v2_type_physical(self):
        physdevs = [
            ["aa:bb:cc:dd:ee:ff", "eth0", "virtio", "0x1000"],
            ["00:11:22:33:44:55", "ens3", "e1000", "0x1643"],
            ["09:87:65:43:21:10", "ens0p1", "mlx4_core", "0:0:1000"],
        ]
        netcfg = {
            "version": 2,
            "ethernets": {args[1]: _mk_v2_phys(*args) for args in physdevs},
        }
        assert sorted(physdevs) == sorted(net.extract_physdevs(netcfg))

    def test_get_v2_type_physical_skips_if_no_set_name(self):
        netcfg = {
            "version": 2,
            "ethernets": {
                "ens3": {
                    "match": {"macaddress": "00:11:22:33:44:55"},
                }
            },
        }
        assert [] == net.extract_physdevs(netcfg)

    def test_runtime_error_on_unknown_netcfg_version(self):
        with pytest.raises(RuntimeError):
            net.extract_physdevs({"version": 3, "awesome_config": []})


class TestNetFailOver:
    @pytest.fixture(autouse=True)
    def setup(self, mocker):
        mocker.patch("cloudinit.net.util")
        self.device_driver = mocker.patch("cloudinit.net.device_driver")
        self.read_sys_net = mocker.patch("cloudinit.net.read_sys_net")

    def test_get_dev_features(self):
        devname = random_string()
        features = random_string()
        self.read_sys_net.return_value = features

        assert features == net.get_dev_features(devname)
        assert 1 == self.read_sys_net.call_count
        self.read_sys_net.assert_called_once_with(devname, "device/features")

    def test_get_dev_features_none_returns_empty_string(self):
        devname = random_string()
        self.read_sys_net.side_effect = Exception("error")
        assert "" == net.get_dev_features(devname)
        assert 1 == self.read_sys_net.call_count
        self.read_sys_net.assert_called_once_with(devname, "device/features")

    @mock.patch("cloudinit.net.get_dev_features")
    def test_has_netfail_standby_feature(self, m_dev_features):
        devname = random_string()
        standby_features = ("0" * 62) + "1" + "0"
        m_dev_features.return_value = standby_features
        assert net.has_netfail_standby_feature(devname)

    @mock.patch("cloudinit.net.get_dev_features")
    def test_has_netfail_standby_feature_short_is_false(self, m_dev_features):
        devname = random_string()
        standby_features = random_string()
        m_dev_features.return_value = standby_features
        assert not net.has_netfail_standby_feature(devname)

    @mock.patch("cloudinit.net.get_dev_features")
    def test_has_netfail_standby_feature_not_present_is_false(
        self, m_dev_features
    ):
        devname = random_string()
        standby_features = "0" * 64
        m_dev_features.return_value = standby_features
        assert not net.has_netfail_standby_feature(devname)

    @mock.patch("cloudinit.net.get_dev_features")
    def test_has_netfail_standby_feature_no_features_is_false(
        self, m_dev_features
    ):
        devname = random_string()
        standby_features = None
        m_dev_features.return_value = standby_features
        assert not net.has_netfail_standby_feature(devname)

    @mock.patch("cloudinit.net.has_netfail_standby_feature")
    @mock.patch("cloudinit.net.os.path.exists")
    def test_is_netfail_master(self, m_exists, m_standby):
        devname = random_string()
        driver = "virtio_net"
        m_exists.return_value = False  # no master sysfs attr
        m_standby.return_value = True  # has standby feature flag
        assert net.is_netfail_master(devname, driver)

    @mock.patch("cloudinit.net.sys_dev_path")
    def test_is_netfail_master_checks_master_attr(self, m_sysdev):
        devname = random_string()
        driver = "virtio_net"
        m_sysdev.return_value = random_string()
        assert not net.is_netfail_master(devname, driver)
        assert 1 == m_sysdev.call_count
        m_sysdev.assert_called_once_with(devname, path="master")

    @mock.patch("cloudinit.net.has_netfail_standby_feature")
    @mock.patch("cloudinit.net.os.path.exists")
    def test_is_netfail_master_wrong_driver(self, m_exists, m_standby):
        devname = random_string()
        driver = random_string()
        assert not net.is_netfail_master(devname, driver)

    @mock.patch("cloudinit.net.has_netfail_standby_feature")
    @mock.patch("cloudinit.net.os.path.exists")
    def test_is_netfail_master_has_master_attr(self, m_exists, m_standby):
        devname = random_string()
        driver = "virtio_net"
        m_exists.return_value = True  # has master sysfs attr
        assert not net.is_netfail_master(devname, driver)

    @mock.patch("cloudinit.net.has_netfail_standby_feature")
    @mock.patch("cloudinit.net.os.path.exists")
    def test_is_netfail_master_no_standby_feat(self, m_exists, m_standby):
        devname = random_string()
        driver = "virtio_net"
        m_exists.return_value = False  # no master sysfs attr
        m_standby.return_value = False  # no standby feature flag
        assert not net.is_netfail_master(devname, driver)

    @mock.patch("cloudinit.net.has_netfail_standby_feature")
    @mock.patch("cloudinit.net.os.path.exists")
    @mock.patch("cloudinit.net.sys_dev_path")
    def test_is_netfail_primary(self, m_sysdev, m_exists, m_standby):
        devname = random_string()
        driver = random_string()  # device not virtio_net
        master_devname = random_string()
        m_sysdev.return_value = "%s/%s" % (
            random_string(),
            master_devname,
        )
        m_exists.return_value = True  # has master sysfs attr
        self.device_driver.return_value = "virtio_net"  # master virtio_net
        m_standby.return_value = True  # has standby feature flag
        assert net.is_netfail_primary(devname, driver)
        self.device_driver.assert_called_once_with(master_devname)
        assert 1 == m_standby.call_count
        m_standby.assert_called_once_with(master_devname)

    @mock.patch("cloudinit.net.has_netfail_standby_feature")
    @mock.patch("cloudinit.net.os.path.exists")
    @mock.patch("cloudinit.net.sys_dev_path")
    def test_is_netfail_primary_wrong_driver(
        self, m_sysdev, m_exists, m_standby
    ):
        devname = random_string()
        driver = "virtio_net"
        assert not net.is_netfail_primary(devname, driver)

    @mock.patch("cloudinit.net.has_netfail_standby_feature")
    @mock.patch("cloudinit.net.os.path.exists")
    @mock.patch("cloudinit.net.sys_dev_path")
    def test_is_netfail_primary_no_master(self, m_sysdev, m_exists, m_standby):
        devname = random_string()
        driver = random_string()  # device not virtio_net
        m_exists.return_value = False  # no master sysfs attr
        assert not net.is_netfail_primary(devname, driver)

    @mock.patch("cloudinit.net.has_netfail_standby_feature")
    @mock.patch("cloudinit.net.os.path.exists")
    @mock.patch("cloudinit.net.sys_dev_path")
    def test_is_netfail_primary_bad_master(
        self, m_sysdev, m_exists, m_standby
    ):
        devname = random_string()
        driver = random_string()  # device not virtio_net
        master_devname = random_string()
        m_sysdev.return_value = "%s/%s" % (
            random_string(),
            master_devname,
        )
        m_exists.return_value = True  # has master sysfs attr
        self.device_driver.return_value = "XXXX"  # master not virtio_net
        assert not net.is_netfail_primary(devname, driver)

    @mock.patch("cloudinit.net.has_netfail_standby_feature")
    @mock.patch("cloudinit.net.os.path.exists")
    @mock.patch("cloudinit.net.sys_dev_path")
    def test_is_netfail_primary_no_standby(
        self, m_sysdev, m_exists, m_standby
    ):
        devname = random_string()
        driver = random_string()  # device not virtio_net
        master_devname = random_string()
        m_sysdev.return_value = "%s/%s" % (
            random_string(),
            master_devname,
        )
        m_exists.return_value = True  # has master sysfs attr
        self.device_driver.return_value = "virtio_net"  # master virtio_net
        m_standby.return_value = False  # master has no standby feature flag
        assert not net.is_netfail_primary(devname, driver)

    @mock.patch("cloudinit.net.has_netfail_standby_feature")
    @mock.patch("cloudinit.net.os.path.exists")
    def test_is_netfail_standby(self, m_exists, m_standby):
        devname = random_string()
        driver = "virtio_net"
        m_exists.return_value = True  # has master sysfs attr
        m_standby.return_value = True  # has standby feature flag
        assert net.is_netfail_standby(devname, driver)

    @mock.patch("cloudinit.net.has_netfail_standby_feature")
    @mock.patch("cloudinit.net.os.path.exists")
    def test_is_netfail_standby_wrong_driver(self, m_exists, m_standby):
        devname = random_string()
        driver = random_string()
        assert not net.is_netfail_standby(devname, driver)

    @mock.patch("cloudinit.net.has_netfail_standby_feature")
    @mock.patch("cloudinit.net.os.path.exists")
    def test_is_netfail_standby_no_master(self, m_exists, m_standby):
        devname = random_string()
        driver = "virtio_net"
        m_exists.return_value = False  # has master sysfs attr
        assert not net.is_netfail_standby(devname, driver)

    @mock.patch("cloudinit.net.has_netfail_standby_feature")
    @mock.patch("cloudinit.net.os.path.exists")
    def test_is_netfail_standby_no_standby_feature(self, m_exists, m_standby):
        devname = random_string()
        driver = "virtio_net"
        m_exists.return_value = True  # has master sysfs attr
        m_standby.return_value = False  # has standby feature flag
        assert not net.is_netfail_standby(devname, driver)

    @mock.patch("cloudinit.net.is_netfail_standby")
    @mock.patch("cloudinit.net.is_netfail_primary")
    def test_is_netfailover_primary(self, m_primary, m_standby):
        devname = random_string()
        driver = random_string()
        m_primary.return_value = True
        m_standby.return_value = False
        assert net.is_netfailover(devname, driver)

    @mock.patch("cloudinit.net.is_netfail_standby")
    @mock.patch("cloudinit.net.is_netfail_primary")
    def test_is_netfailover_standby(self, m_primary, m_standby):
        devname = random_string()
        driver = random_string()
        m_primary.return_value = False
        m_standby.return_value = True
        assert net.is_netfailover(devname, driver)

    @mock.patch("cloudinit.net.is_netfail_standby")
    @mock.patch("cloudinit.net.is_netfail_primary")
    def test_is_netfailover_returns_false(self, m_primary, m_standby):
        devname = random_string()
        driver = random_string()
        m_primary.return_value = False
        m_standby.return_value = False
        assert not net.is_netfailover(devname, driver)


class TestOpenvswitchIsInstalled:
    """Test cloudinit.net.openvswitch_is_installed.

    Uses the ``clear_lru_cache`` local autouse fixture to allow us to test
    despite the ``lru_cache`` decorator on the unit under test.
    """

    @pytest.fixture(autouse=True)
    def clear_lru_cache(self):
        net.openvswitch_is_installed.cache_clear()

    @pytest.mark.parametrize(
        "expected,which_return", [(True, "/some/path"), (False, None)]
    )
    @mock.patch("cloudinit.net.subp.which")
    def test_mirrors_which_result(self, m_which, expected, which_return):
        m_which.return_value = which_return
        assert expected == net.openvswitch_is_installed()

    @mock.patch("cloudinit.net.subp.which")
    def test_only_calls_which_once(self, m_which):
        net.openvswitch_is_installed()
        net.openvswitch_is_installed()
        assert 1 == m_which.call_count


@mock.patch("cloudinit.net.subp.subp", return_value=("", ""))
class TestGetOVSInternalInterfaces:
    """Test cloudinit.net.get_ovs_internal_interfaces.

    Uses the ``clear_lru_cache`` local autouse fixture to allow us to test
    despite the ``lru_cache`` decorator on the unit under test.
    """

    @pytest.fixture(autouse=True)
    def clear_lru_cache(self):
        net.get_ovs_internal_interfaces.cache_clear()

    def test_command_used(self, m_subp):
        """Test we use the correct command when we call subp"""
        net.get_ovs_internal_interfaces()

        assert [
            mock.call(net.OVS_INTERNAL_INTERFACE_LOOKUP_CMD)
        ] == m_subp.call_args_list

    def test_subp_contents_split_and_returned(self, m_subp):
        """Test that the command output is appropriately mangled."""
        stdout = "iface1\niface2\niface3\n"
        m_subp.return_value = (stdout, "")

        assert [
            "iface1",
            "iface2",
            "iface3",
        ] == net.get_ovs_internal_interfaces()

    def test_database_connection_error_handled_gracefully(self, m_subp):
        """Test that the error indicating OVS is down is handled gracefully."""
        m_subp.side_effect = ProcessExecutionError(
            stderr="database connection failed"
        )

        assert [] == net.get_ovs_internal_interfaces()

    def test_other_errors_raised(self, m_subp):
        """Test that only database connection errors are handled."""
        m_subp.side_effect = ProcessExecutionError()

        with pytest.raises(ProcessExecutionError):
            net.get_ovs_internal_interfaces()

    def test_only_runs_once(self, m_subp):
        """Test that we cache the value."""
        net.get_ovs_internal_interfaces()
        net.get_ovs_internal_interfaces()

        assert 1 == m_subp.call_count


@mock.patch("cloudinit.net.get_ovs_internal_interfaces")
@mock.patch("cloudinit.net.openvswitch_is_installed")
class TestIsOpenVSwitchInternalInterface:
    def test_false_if_ovs_not_installed(
        self, m_openvswitch_is_installed, _m_get_ovs_internal_interfaces
    ):
        """Test that OVS' absence returns False."""
        m_openvswitch_is_installed.return_value = False

        assert not net.is_openvswitch_internal_interface("devname")

    @pytest.mark.parametrize(
        "detected_interfaces,devname,expected_return",
        [
            ([], "devname", False),
            (["notdevname"], "devname", False),
            (["devname"], "devname", True),
            (["some", "other", "devices", "and", "ours"], "ours", True),
        ],
    )
    def test_return_value_based_on_detected_interfaces(
        self,
        m_openvswitch_is_installed,
        m_get_ovs_internal_interfaces,
        detected_interfaces,
        devname,
        expected_return,
    ):
        """Test that the detected interfaces are used correctly."""
        m_openvswitch_is_installed.return_value = True
        m_get_ovs_internal_interfaces.return_value = detected_interfaces
        assert expected_return == net.is_openvswitch_internal_interface(
            devname
        )


class TestIsIpAddress:
    """Tests for net.is_ip_address.

    Instead of testing with values we rely on the ipaddress stdlib module to
    handle all values correctly, so simply test that is_ip_address defers to
    the ipaddress module correctly.
    """

    @pytest.mark.parametrize(
        "ip_address_side_effect,expected_return",
        (
            (ValueError, False),
            (lambda _: ipaddress.IPv4Address("192.168.0.1"), True),
            (lambda _: ipaddress.IPv4Address("192.168.0.1/24"), False),
            (lambda _: ipaddress.IPv6Address("2001:db8::"), True),
            (lambda _: ipaddress.IPv6Address("2001:db8::/48"), False),
        ),
    )
    def test_is_ip_address(self, ip_address_side_effect, expected_return):
        with mock.patch(
            "cloudinit.net.ipaddress.ip_address",
            side_effect=ip_address_side_effect,
        ) as m_ip_address:
            ret = net.is_ip_address(mock.sentinel.ip_address_in)
        assert expected_return == ret
        expected_call = mock.call(mock.sentinel.ip_address_in)
        assert [expected_call] == m_ip_address.call_args_list


class TestIsIpv4Address:
    """Tests for net.is_ipv4_address.

    Instead of testing with values we rely on the ipaddress stdlib module to
    handle all values correctly, so simply test that is_ipv4_address defers to
    the ipaddress module correctly.
    """

    @pytest.mark.parametrize(
        "ipv4address_mock,expected_return",
        (
            (mock.Mock(side_effect=ValueError), False),
            (
                mock.Mock(return_value=ipaddress.IPv4Address("192.168.0.1")),
                True,
            ),
        ),
    )
    def test_is_ip_address(self, ipv4address_mock, expected_return):
        with mock.patch(
            "cloudinit.net.ipaddress.IPv4Address", ipv4address_mock
        ) as m_ipv4address:
            ret = net.is_ipv4_address(mock.sentinel.ip_address_in)
        assert expected_return == ret
        expected_call = mock.call(mock.sentinel.ip_address_in)
        assert [expected_call] == m_ipv4address.call_args_list


class TestIsIpNetwork:
    """Tests for net.is_ip_network() and related functions."""

    @pytest.mark.parametrize(
        "func,arg,expected_return",
        (
            (net.is_ip_network, "192.168.1.1", True),
            (net.is_ip_network, "192.168.1.1/24", True),
            (net.is_ip_network, "192.168.1.1/32", True),
            (net.is_ip_network, "192.168.1.1/33", False),
            (net.is_ip_network, "2001:67c:1", False),
            (net.is_ip_network, "2001:67c:1/32", False),
            (net.is_ip_network, "2001:67c::", True),
            (net.is_ip_network, "2001:67c::/32", True),
            (net.is_ipv4_network, "192.168.1.1", True),
            (net.is_ipv4_network, "192.168.1.1/24", True),
            (net.is_ipv4_network, "2001:67c::", False),
            (net.is_ipv4_network, "2001:67c::/32", False),
            (net.is_ipv6_network, "192.168.1.1", False),
            (net.is_ipv6_network, "192.168.1.1/24", False),
            (net.is_ipv6_network, "2001:67c:1", False),
            (net.is_ipv6_network, "2001:67c:1/32", False),
            (net.is_ipv6_network, "2001:67c::", True),
            (net.is_ipv6_network, "2001:67c::/32", True),
            (net.is_ipv6_network, "2001:67c::/129", False),
            (net.is_ipv6_network, "2001:67c::/128", True),
        ),
    )
    def test_is_ip_network(self, func, arg, expected_return):
        assert func(arg) == expected_return


class TestIsIpInSubnet:
    """Tests for net.is_ip_in_subnet()."""

    @pytest.mark.parametrize(
        "func,ip,subnet,expected_return",
        (
            (net.is_ip_in_subnet, "192.168.1.1", "2001:67c::1/64", False),
            (net.is_ip_in_subnet, "2001:67c::1", "192.168.1.1/24", False),
            (net.is_ip_in_subnet, "192.168.1.1", "192.168.1.1/24", True),
            (net.is_ip_in_subnet, "192.168.1.1", "192.168.1.1/32", True),
            (net.is_ip_in_subnet, "192.168.1.2", "192.168.1.1/24", True),
            (net.is_ip_in_subnet, "192.168.1.2", "192.168.1.1/32", False),
            (net.is_ip_in_subnet, "192.168.2.2", "192.168.1.1/24", False),
            (net.is_ip_in_subnet, "192.168.2.2", "192.168.1.1/32", False),
            (net.is_ip_in_subnet, "2001:67c1::1", "2001:67c1::1/64", True),
            (net.is_ip_in_subnet, "2001:67c1::1", "2001:67c1::1/128", True),
            (net.is_ip_in_subnet, "2001:67c1::2", "2001:67c1::1/64", True),
            (net.is_ip_in_subnet, "2001:67c1::2", "2001:67c1::1/128", False),
            (net.is_ip_in_subnet, "2002:67c1::1", "2001:67c1::1/8", True),
            (net.is_ip_in_subnet, "2002:67c1::1", "2001:67c1::1/16", False),
        ),
    )
    def test_is_ip_in_subnet(self, func, ip, subnet, expected_return):
        assert func(ip, subnet) == expected_return



# This file is part of cloud-init. See LICENSE file for license information.

import functools
import os
from unittest import mock

import pytest
import responses

from cloudinit.sources import DataSourceAliYun as ay
from cloudinit.sources.helpers.aliyun import (
    convert_ecs_metadata_network_config,
)
from cloudinit.util import load_json

DEFAULT_METADATA_RAW = r"""{
  "disks": {
    "bp15spwwhlf8bbbn7xxx": {
      "id": "d-bp15spwwhlf8bbbn7xxx",
      "name": ""
    }
  },
  "dns-conf": {
    "nameservers": [
    "100.100.2.136",
    "100.100.2.138"
    ]
  },
  "hibernation": {
    "configured": "false"
  },
  "instance": {
    "instance-name": "aliyun-test-vm-00",
    "instance-type": "ecs.g8i.large",
    "last-host-landing-time": "2024-11-17 10:02:41",
    "max-netbw-egress": "2560000",
    "max-netbw-ingress": "2560000",
    "virtualization-solution": "ECS Virt",
    "virtualization-solution-version": "2.0"
  },
  "network": {
    "interfaces": {
      "macs": {
        "00:16:3e:14:59:58": {
          "gateway": "172.16.101.253",
          "netmask": "255.255.255.0",
          "network-interface-id": "eni-bp13i3ed90icgdgaxxxx"
        }
      }
    }
  },
  "ntp-conf": {
    "ntp-servers": [
      "ntp1.aliyun.com",
      "ntp1.cloud.aliyuncs.com"
    ]
  },
  "public-keys": {
    "0": {
      "openssh-key": "ssh-rsa AAAAB3Nza"
    },
    "skp-bp1test": {
      "openssh-key": "ssh-rsa AAAAB3Nza"
    }
  },
  "eipv4": "121.66.77.88",
  "hostname": "aliyun-test-vm-00",
  "image-id": "ubuntu_24_04_x64_20G_alibase_20241016.vhd",
  "instance-id": "i-bp15ojxppkmsnyjxxxxx",
  "mac": "00:16:3e:14:59:58",
  "network-type": "vpc",
  "owner-account-id": "123456",
  "private-ipv4": "172.16.111.222",
  "region-id": "cn-hangzhou",
  "serial-number": "3ca05955-a892-46b3-a6fc-xxxxxx",
  "source-address": "http://mirrors.cloud.aliyuncs.com",
  "sub-private-ipv4-list": "172.16.101.215",
  "vpc-cidr-block": "172.16.0.0/12",
  "vpc-id": "vpc-bp1uwvjta7txxxxxxx",
  "vswitch-cidr-block": "172.16.101.0/24",
  "vswitch-id": "vsw-bp12cibmw6078qv123456",
  "zone-id": "cn-hangzhou-j"
}"""

DEFAULT_METADATA = load_json(DEFAULT_METADATA_RAW)

DEFAULT_USERDATA = """\
#cloud-config

hostname: localhost"""

DEFAULT_VENDORDATA = """\
#cloud-config
bootcmd:
- echo hello world > /tmp/vendor"""


@pytest.fixture
def ds(paths):
    cfg = {"datasource": {"AliYun": {"timeout": "1", "max_wait": "1"}}}
    distro = {}
    return ay.DataSourceAliYun(cfg, distro, paths)


@pytest.fixture
def metadata_address(ds):
    return ds.metadata_urls[0]


def register_mock_metaserver(base_url, data):
    def register_helper(register, base_url, body):
        if isinstance(body, str):
            register(base_url, body)
        elif isinstance(body, list):
            register(base_url.rstrip("/"), "\n".join(body) + "\n")
        elif isinstance(body, dict):
            if not body:
                register(base_url.rstrip("/") + "/", "not found", status=404)
            vals = []
            for k, v in body.items():
                if isinstance(v, (str, list)):
                    suffix = k.rstrip("/")
                else:
                    suffix = k.rstrip("/") + "/"
                vals.append(suffix)
                url = base_url.rstrip("/") + "/" + suffix
                register_helper(register, url, v)
            register(base_url, "\n".join(vals) + "\n")

    register = functools.partial(responses.add, responses.GET)
    register_helper(register, base_url, data)


@pytest.fixture
def regist_default_server(ds, metadata_address):
    metadata_url = (
        os.path.join(
            metadata_address,
            ds.min_metadata_version,
            "meta-data",
        )
        + "/"
    )
    register_mock_metaserver(metadata_url, DEFAULT_METADATA)

    userdata_url = os.path.join(
        metadata_address, ds.min_metadata_version, "user-data"
    )
    register_mock_metaserver(userdata_url, DEFAULT_USERDATA)

    vendordata_url = os.path.join(
        metadata_address, ds.min_metadata_version, "vendor-data"
    )
    register_mock_metaserver(vendordata_url, DEFAULT_USERDATA)

    # EC2 provides an instance-identity document which must return 404 here
    # for this test to pass.
    default_identity = {}
    identity_url = os.path.join(
        metadata_address,
        ds.min_metadata_version,
        "dynamic",
        "instance-identity",
    )
    register_mock_metaserver(identity_url, default_identity)

    token_url = os.path.join(metadata_address, "latest", "api", "token")
    responses.add(responses.PUT, token_url, "API-TOKEN")


@pytest.fixture
def regist_json_meta_path(ds, metadata_address):
    metadata_all_url = (
        os.path.join(
            metadata_address,
            ds.min_metadata_version,
            "meta-data",
        )
        + "/all"
    )
    register_mock_metaserver(metadata_all_url, DEFAULT_METADATA_RAW)


class TestAliYunDatasource:

    def _test_get_data(self, ds):
        assert ds.metadata == DEFAULT_METADATA
        assert ds.userdata_raw == DEFAULT_USERDATA.encode("utf8")

    def _test_get_sshkey(self, ds):
        pub_keys = [
            v["openssh-key"]
            for (_, v) in DEFAULT_METADATA["public-keys"].items()
        ]
        assert ds.get_public_ssh_keys() == pub_keys

    def _test_get_iid(self, ds):
        assert DEFAULT_METADATA["instance-id"] == ds.get_instance_id()

    def _test_host_name(self, ds):
        assert DEFAULT_METADATA["hostname"] == ds.get_hostname().hostname

    @responses.activate
    @pytest.mark.usefixtures("regist_default_server", "regist_json_meta_path")
    @mock.patch("cloudinit.sources.DataSourceEc2.util.is_resolvable")
    @mock.patch("cloudinit.sources.DataSourceAliYun._is_aliyun")
    def test_with_mock_server(self, m_is_aliyun, m_resolv, ds):
        m_is_aliyun.return_value = True
        ret = ds.get_data()
        assert True is ret
        assert 1 == m_is_aliyun.call_count
        self._test_get_data(ds)
        self._test_get_sshkey(ds)
        self._test_get_iid(ds)
        self._test_host_name(ds)
        assert "aliyun" == ds.cloud_name
        assert "aliyun" == ds.platform
        assert "metadata (http://100.100.100.200)" == ds.subplatform

    @responses.activate
    @pytest.mark.usefixtures("regist_default_server")
    @mock.patch("cloudinit.sources.DataSourceEc2.util.is_resolvable")
    @mock.patch("cloudinit.sources.DataSourceAliYun._is_aliyun")
    def test_with_mock_server_without_json_path(
        self, m_is_aliyun, m_resolv, ds
    ):
        m_is_aliyun.return_value = True
        ret = ds.get_data()
        assert True is ret
        assert 1 == m_is_aliyun.call_count
        self._test_get_data(ds)
        self._test_get_sshkey(ds)
        self._test_get_iid(ds)
        self._test_host_name(ds)
        assert "aliyun" == ds.cloud_name
        assert "aliyun" == ds.platform
        assert "metadata (http://100.100.100.200)" == ds.subplatform

    @responses.activate
    @pytest.mark.usefixtures("regist_default_server", "regist_json_meta_path")
    @mock.patch("cloudinit.net.ephemeral.EphemeralIPv6Network")
    @mock.patch("cloudinit.net.ephemeral.EphemeralIPv4Network")
    @mock.patch("cloudinit.sources.DataSourceEc2.util.is_resolvable")
    @mock.patch("cloudinit.sources.DataSourceAliYun._is_aliyun")
    @mock.patch("cloudinit.net.find_fallback_nic")
    @mock.patch("cloudinit.net.ephemeral.maybe_perform_dhcp_discovery")
    @mock.patch("cloudinit.sources.DataSourceEc2.util.is_FreeBSD")
    @pytest.mark.usefixtures("disable_netdev_info")
    def test_aliyun_local_with_mock_server(
        self,
        m_is_bsd,
        m_dhcp,
        m_fallback_nic,
        m_is_aliyun,
        m_resolva,
        m_net4,
        m_net6,
        ds,
        paths,
    ):
        m_is_aliyun.return_value = True
        m_fallback_nic.return_value = "eth9"
        m_dhcp.return_value = {
            "interface": "eth9",
            "fixed-address": "192.168.2.9",
            "routers": "192.168.2.1",
            "subnet-mask": "255.255.255.0",
            "broadcast-address": "192.168.2.255",
        }
        m_is_bsd.return_value = False
        cfg = {"datasource": {"AliYun": {"timeout": "1", "max_wait": "1"}}}
        distro = mock.MagicMock()
        ds = ay.DataSourceAliYunLocal(cfg, distro, paths)
        ret = ds.get_data()
        assert True is ret
        assert 1 == m_is_aliyun.call_count
        self._test_get_data(ds)
        self._test_get_sshkey(ds)
        self._test_get_iid(ds)
        self._test_host_name(ds)
        assert "aliyun" == ds.cloud_name
        assert "aliyun" == ds.platform
        assert "metadata (http://100.100.100.200)" == ds.subplatform

    @responses.activate
    @pytest.mark.usefixtures("regist_default_server", "regist_json_meta_path")
    @mock.patch("cloudinit.sources.DataSourceAliYun._is_aliyun")
    def test_returns_false_when_not_on_aliyun(self, m_is_aliyun, ds):
        """If is_aliyun returns false, then get_data should return False."""
        m_is_aliyun.return_value = False
        ret = ds.get_data()
        assert 1 == m_is_aliyun.call_count
        assert False is ret

    def test_parse_public_keys(self):
        public_keys = {}
        assert ay.parse_public_keys(public_keys) == []

        public_keys = {"key-pair-0": "ssh-key-0"}
        assert ay.parse_public_keys(public_keys) == [public_keys["key-pair-0"]]

        public_keys = {"key-pair-0": "ssh-key-0", "key-pair-1": "ssh-key-1"}
        assert set(ay.parse_public_keys(public_keys)) == set(
            [public_keys["key-pair-0"], public_keys["key-pair-1"]]
        )

        public_keys = {"key-pair-0": ["ssh-key-0", "ssh-key-1"]}
        assert ay.parse_public_keys(public_keys) == public_keys["key-pair-0"]

        public_keys = {"key-pair-0": {"openssh-key": []}}
        assert ay.parse_public_keys(public_keys) == []

        public_keys = {"key-pair-0": {"openssh-key": "ssh-key-0"}}
        assert ay.parse_public_keys(public_keys) == [
            public_keys["key-pair-0"]["openssh-key"]
        ]

        public_keys = {
            "key-pair-0": {"openssh-key": ["ssh-key-0", "ssh-key-1"]}
        }
        assert (
            ay.parse_public_keys(public_keys)
            == public_keys["key-pair-0"]["openssh-key"]
        )

    def test_route_metric_calculated_with_multiple_network_cards(self):
        """Test that route-metric code works with multiple network cards"""
        netcfg = convert_ecs_metadata_network_config(
            {
                "interfaces": {
                    "macs": {
                        "00:16:3e:14:59:58": {
                            "ipv6-gateway": "2408:xxxxx",
                            "ipv6s": "[2408:xxxxxx]",
                            "network-interface-id": "eni-bp13i1xxxxx",
                        },
                        "00:16:3e:39:43:27": {
                            "gateway": "172.16.101.253",
                            "netmask": "255.255.255.0",
                            "network-interface-id": "eni-bp13i2xxxx",
                        },
                    }
                }
            },
            macs_to_nics={
                "00:16:3e:14:59:58": "eth0",
                "00:16:3e:39:43:27": "eth1",
            },
        )

        met0 = netcfg["ethernets"]["eth0"]["dhcp4-overrides"]["route-metric"]
        met1 = netcfg["ethernets"]["eth1"]["dhcp4-overrides"]["route-metric"]

        # route-metric numbers should be 100 apart
        assert 100 == abs(met0 - met1)

        # No policy routing
        assert not {"routing-policy", "routes"}.intersection(
            netcfg["ethernets"]["eth0"].keys()
        )
        assert not {"routing-policy", "routes"}.intersection(
            netcfg["ethernets"]["eth1"].keys()
        )

        # eth0 network meta-data  have ipv6s info, ipv6 should True
        met0_dhcp6 = netcfg["ethernets"]["eth0"]["dhcp6"]
        assert met0_dhcp6 is True

        netcfg = convert_ecs_metadata_network_config(
            {
                "interfaces": {
                    "macs": {
                        "00:16:3e:14:59:58": {
                            "gateway": "172.16.101.253",
                            "netmask": "255.255.255.0",
                            "network-interface-id": "eni-bp13ixxxx",
                        }
                    }
                }
            },
            macs_to_nics={"00:16:3e:14:59:58": "eth0"},
        )
        met0 = netcfg["ethernets"]["eth0"]
        # single network card would have no dhcp4-overrides
        assert "dhcp4-overrides" not in met0


class TestIsAliYun:
    ALIYUN_PRODUCT = "Alibaba Cloud ECS"
    read_dmi_data_expected = [mock.call("system-product-name")]

    @mock.patch("cloudinit.sources.DataSourceAliYun.dmi.read_dmi_data")
    def test_true_on_aliyun_product(self, m_read_dmi_data):
        """Should return true if the dmi product data has expected value."""
        m_read_dmi_data.return_value = self.ALIYUN_PRODUCT
        ret = ay._is_aliyun()
        assert self.read_dmi_data_expected == m_read_dmi_data.call_args_list
        assert True is ret

    @mock.patch("cloudinit.sources.DataSourceAliYun.dmi.read_dmi_data")
    def test_false_on_empty_string(self, m_read_dmi_data):
        """Should return false on empty value returned."""
        m_read_dmi_data.return_value = ""
        ret = ay._is_aliyun()
        assert self.read_dmi_data_expected == m_read_dmi_data.call_args_list
        assert False is ret

    @mock.patch("cloudinit.sources.DataSourceAliYun.dmi.read_dmi_data")
    def test_false_on_unknown_string(self, m_read_dmi_data):
        """Should return false on an unrelated string."""
        m_read_dmi_data.return_value = "cubs win"
        ret = ay._is_aliyun()
        assert self.read_dmi_data_expected == m_read_dmi_data.call_args_list
        assert False is ret



import os
from unittest import mock

import pytest


@pytest.fixture(autouse=True)
def mock_util_get_cmdline():
    with mock.patch("cloudinit.util.get_cmdline", return_value="") as m:
        yield m


@pytest.fixture(autouse=True)
def hide_resource_disk():
    """GitHub runner may have a resource disk which may affect tests."""
    real_exists = os.path.exists
    with mock.patch("os.path.exists") as mock_exists:
        mock_exists.side_effect = lambda path: (
            False
            if path == "/dev/disk/cloud/azure_resource"
            else real_exists(path)
        )
        yield



# Copyright (C) 2015 Canonical Ltd.
#
# Author: Ryan Harper <ryan.harper@canonical.com>
#
# This file is part of cloud-init. See LICENSE file for license information.


def compose_udev_equality(key, value):
    """Return a udev comparison clause, like `ACTION=="add"`."""
    assert key == key.upper()
    return '%s=="%s"' % (key, value)


def compose_udev_attr_equality(attribute, value):
    """Return a udev attribute comparison clause, like `ATTR{type}=="1"`."""
    assert attribute == attribute.lower()
    return 'ATTR{%s}=="%s"' % (attribute, value)


def compose_udev_setting(key, value):
    """Return a udev assignment clause, like `NAME="eth0"`."""
    assert key == key.upper()
    return '%s="%s"' % (key, value)


def generate_udev_rule(interface, mac, driver=None):
    """Return a udev rule to set the name of network interface with `mac`.

    The rule ends up as a single line looking something like:

    SUBSYSTEM=="net", ACTION=="add", DRIVERS=="?*",
    ATTR{address}="ff:ee:dd:cc:bb:aa", NAME="eth0"
    """
    if not driver:
        driver = "?*"

    rule = ", ".join(
        [
            compose_udev_equality("SUBSYSTEM", "net"),
            compose_udev_equality("ACTION", "add"),
            compose_udev_equality("DRIVERS", driver),
            compose_udev_attr_equality("address", mac),
            compose_udev_setting("NAME", interface),
        ]
    )
    return "%s\n" % rule



# This file is part of cloud-init. See LICENSE file for license information.

import os

import pytest

from cloudinit import safeyaml, stages, util
from cloudinit.config.modules import Modules
from cloudinit.settings import PER_INSTANCE
from tests.unittests import helpers
from tests.unittests.helpers import replicate_test_root


@pytest.fixture(autouse=True)
def user_data(tmp_path):
    replicate_test_root("simple_ubuntu", str(tmp_path))
    return helpers.readResource("user_data.1.txt")


@pytest.fixture(autouse=True)
def cfg(tmp_path, mocker):
    # root group doesn't exist everywhere
    mocker.patch("cloudinit.util.chownbyname")

    new_root = str(tmp_path)
    cfg = {
        "datasource_list": ["None"],
        "cloud_init_modules": ["write_files"],
        "system_info": {
            "paths": {"run_dir": new_root},
            "package_mirrors": [
                {
                    "arches": ["i386", "amd64", "blah"],
                    "failsafe": {
                        "primary": "http://my.archive.mydomain.com/ubuntu",
                        "security": ("http://my.security.mydomain.com/ubuntu"),
                    },
                    "search": {"primary": [], "security": []},
                },
            ],
        },
    }
    cloud_cfg = safeyaml.dumps(cfg)
    util.ensure_dir(os.path.join(new_root, "etc", "cloud"))
    util.write_file(
        os.path.join(new_root, "etc", "cloud", "cloud.cfg"), cloud_cfg
    )


@pytest.mark.usefixtures("fake_filesystem")
class TestMergeRun:
    def test_none_ds(self, user_data):
        initer = stages.Init()
        initer.read_cfg()
        initer.initialize()
        initer.fetch()
        initer.datasource.userdata_raw = user_data
        initer.instancify()
        initer.update()
        initer.cloudify().run(
            "consume_data",
            initer.consume_data,
            args=[PER_INSTANCE],
            freq=PER_INSTANCE,
        )
        mirrors = initer.distro.get_option("package_mirrors")
        assert 1 == len(mirrors)
        mirror = mirrors[0]
        assert mirror["arches"] == ["i386", "amd64", "blah"]
        mods = Modules(initer)
        (which_ran, failures) = mods.run_section("cloud_init_modules")
        assert not failures
        assert os.path.exists("/etc/blah.ini")
        assert "write_files" in which_ran
        contents = util.load_text_file("/etc/blah.ini")
        assert contents == "blah"



import logging
import os

from cloudinit import sources

LOG = logging.getLogger(__name__)


class DataSourceNoCacheWithFallback(sources.DataSource):
    def _get_data(self):
        if os.path.exists("/ci-test-firstboot"):
            LOG.debug("TEST _get_data called")
            return True
        return False

    def check_if_fallback_is_allowed(self):
        return True


datasources = [
    (
        DataSourceNoCacheWithFallback,
        (sources.DEP_FILESYSTEM,),
    ),
]


def get_datasource_list(depends):
    return sources.list_from_depends(depends, datasources)



# This file is part of cloud-init. See LICENSE file for license information.

import copy
import os
import re

import pytest

from cloudinit import util
from cloudinit.config import cc_apt_configure


@pytest.fixture
def p_c_files(tmp_path):
    pfile = str(tmp_path / "proxy.cfg")
    cfile = str(tmp_path / "config.cfg")
    return pfile, cfile


def _search_apt_config(contents, ptype, value):
    return re.search(
        r"acquire::%s::proxy\s+[\"']%s[\"'];\n" % (ptype, value),
        contents,
        flags=re.IGNORECASE,
    )


class TestAptProxyConfig:
    @pytest.mark.parametrize(
        "cfg", [{"proxy": "myproxy"}, {"http_proxy": "myproxy"}]
    )
    def test_apt_proxy_written_variants(self, p_c_files, cfg):
        pfile, cfile = p_c_files
        cc_apt_configure.apply_apt_config(cfg, pfile, cfile)

        assert os.path.isfile(pfile)
        assert not os.path.isfile(cfile)

        contents = util.load_text_file(pfile)
        assert _search_apt_config(contents, "http", "myproxy")

    def test_apt_all_proxy_written(self, p_c_files):
        pfile, cfile = p_c_files
        cfg = {
            "http_proxy": "myproxy_http_proxy",
            "https_proxy": "myproxy_https_proxy",
            "ftp_proxy": "myproxy_ftp_proxy",
        }

        values = {
            "http": cfg["http_proxy"],
            "https": cfg["https_proxy"],
            "ftp": cfg["ftp_proxy"],
        }

        cc_apt_configure.apply_apt_config(cfg, pfile, cfile)

        assert os.path.isfile(pfile)
        assert not os.path.isfile(cfile)

        contents = util.load_text_file(pfile)

        for ptype, pval in values.items():
            assert _search_apt_config(contents, ptype, pval)

    def test_proxy_deleted(self, p_c_files):
        pfile, cfile = p_c_files
        util.write_file(cfile, "content doesn't matter")
        cc_apt_configure.apply_apt_config({}, pfile, cfile)
        assert not os.path.isfile(pfile)
        assert not os.path.isfile(cfile)

    def test_proxy_replaced(self, p_c_files):
        pfile, cfile = p_c_files
        util.write_file(cfile, "content doesn't matter")
        cc_apt_configure.apply_apt_config({"proxy": "foo"}, pfile, cfile)
        assert os.path.isfile(pfile)
        contents = util.load_text_file(pfile)
        assert _search_apt_config(contents, "http", "foo")

    def test_config_written(self, p_c_files):
        pfile, cfile = p_c_files
        payload = "this is my apt config"
        cfg = {"conf": payload}

        cc_apt_configure.apply_apt_config(cfg, pfile, cfile)

        assert os.path.isfile(cfile)
        assert not os.path.isfile(pfile)

        assert util.load_text_file(cfile) == payload

    def test_config_replaced(self, p_c_files):
        pfile, cfile = p_c_files
        util.write_file(pfile, "content doesn't matter")
        cc_apt_configure.apply_apt_config({"conf": "foo"}, pfile, cfile)
        assert os.path.isfile(cfile)
        assert util.load_text_file(cfile) == "foo"

    def test_config_deleted(self, p_c_files):
        # if no 'conf' is provided, delete any previously written file
        pfile, cfile = p_c_files
        util.write_file(pfile, "content doesn't matter")
        cc_apt_configure.apply_apt_config({}, pfile, cfile)
        assert not os.path.isfile(pfile)
        assert not os.path.isfile(cfile)


class TestConversion:
    def test_convert_with_apt_mirror_as_empty_string(self):
        # an empty apt_mirror is the same as no apt_mirror
        empty_m_found = cc_apt_configure.convert_to_v3_apt_format(
            {"apt_mirror": ""}
        )
        default_found = cc_apt_configure.convert_to_v3_apt_format({})
        assert default_found == empty_m_found

    def test_convert_with_apt_mirror(self):
        mirror = "http://my.mirror/ubuntu"
        f = cc_apt_configure.convert_to_v3_apt_format({"apt_mirror": mirror})
        assert mirror in set(m["uri"] for m in f["apt"]["primary"])

    def test_no_old_content(self):
        mirror = "http://my.mirror/ubuntu"
        mydata = {"apt": {"primary": {"arches": ["default"], "uri": mirror}}}
        expected = copy.deepcopy(mydata)
        assert expected == cc_apt_configure.convert_to_v3_apt_format(mydata)



# Copyright (C) 2014 Vaidas Jablonskis
#
# Author: Vaidas Jablonskis <jablonskis@gmail.com>
#
# This file is part of cloud-init. See LICENSE file for license information.
# pylint: disable=attribute-defined-outside-init

import datetime
import json
import re
from base64 import b64decode, b64encode
from unittest import mock
from urllib.parse import urlparse

import pytest
import responses

from cloudinit import distros, helpers, settings
from cloudinit.net.dhcp import NoDHCPLeaseError
from cloudinit.sources import DataSourceGCE

M_PATH = "cloudinit.sources.DataSourceGCE."

GCE_META = {
    "instance/id": "123",
    "instance/zone": "foo/bar",
    "instance/hostname": "server.project-foo.local",
}

GCE_META_PARTIAL = {
    "instance/id": "1234",
    "instance/hostname": "server.project-bar.local",
    "instance/zone": "bar/baz",
}

GCE_META_ENCODING = {
    "instance/id": "12345",
    "instance/hostname": "server.project-baz.local",
    "instance/zone": "baz/bang",
    "instance/attributes": {
        "user-data": b64encode(b"#!/bin/echo baz\n").decode("utf-8"),
        "user-data-encoding": "base64",
    },
}

GCE_USER_DATA_TEXT = {
    "instance/id": "12345",
    "instance/hostname": "server.project-baz.local",
    "instance/zone": "baz/bang",
    "instance/attributes": {
        "user-data": "#!/bin/sh\necho hi mom\ntouch /run/up-now\n",
    },
}

HEADERS = {"Metadata-Flavor": "Google"}
MD_URL_RE = re.compile(
    r"http://metadata.google.internal/computeMetadata/v1/.*"
)
GUEST_ATTRIBUTES_URL = (
    "http://metadata.google.internal/computeMetadata/"
    "v1/instance/guest-attributes/hostkeys/"
)


class TestDataSourceGCE:
    with_logs = True

    def _make_distro(self, dtype, def_user=None):
        cfg = dict(settings.CFG_BUILTIN)
        cfg["system_info"]["distro"] = dtype
        paths = helpers.Paths(cfg["system_info"]["paths"])
        distro_cls = distros.fetch(dtype)
        if def_user:
            cfg["system_info"]["default_user"] = def_user.copy()
        distro = distro_cls(dtype, cfg["system_info"], paths)
        return distro

    @pytest.fixture(autouse=True)
    def fixtures(self, mocker, paths):
        self.ds = DataSourceGCE.DataSourceGCE(
            settings.CFG_BUILTIN, None, paths
        )
        self.m_platform_reports_gce = mocker.patch(
            M_PATH + "platform_reports_gce", return_value=True
        )
        self.m_is_resolvable_url = mocker.patch(
            M_PATH + "util.is_resolvable_url", return_value=True
        )
        mocker.patch("time.sleep")

    def _set_mock_metadata(self, gce_meta=None, *, check_headers=None):
        if gce_meta is None:
            gce_meta = GCE_META

        def _request_callback(request):
            recursive = False
            url_path = urlparse(request.url).path
            if url_path.startswith("/computeMetadata/v1/"):
                path = url_path.split("/computeMetadata/v1/")[1:][0]
                recursive = path.endswith("/")
                path = path.rstrip("/")
            else:
                path = None
            if path in gce_meta:
                response = gce_meta.get(path)
                if recursive:
                    response = json.dumps(response)
                if check_headers is not None:
                    for k in check_headers.keys():
                        assert check_headers[k] == request.headers[k]
                return (200, request.headers, response)
            else:
                return (404, request.headers, "")

        responses.add_callback(
            responses.GET,
            MD_URL_RE,
            callback=_request_callback,
        )

    @responses.activate
    def test_connection(self):
        self._set_mock_metadata(check_headers=HEADERS)
        success = self.ds.get_data()
        assert success

    @responses.activate
    def test_metadata(self):
        # UnicodeDecodeError if set to ds.userdata instead of userdata_raw
        meta = GCE_META.copy()
        meta["instance/attributes/user-data"] = b"/bin/echo \xff\n"

        self._set_mock_metadata()
        self.ds.get_data()

        shostname = GCE_META.get("instance/hostname").split(".")[0]
        assert shostname == self.ds.get_hostname().hostname

        assert GCE_META.get("instance/id") == self.ds.get_instance_id()

        assert (
            GCE_META.get("instance/attributes/user-data")
            == self.ds.get_userdata_raw()
        )

    @responses.activate
    def test_metadata_partial(self):
        """test partial metadata (missing user-data in particular)"""
        self._set_mock_metadata(GCE_META_PARTIAL)
        self.ds.get_data()

        assert GCE_META_PARTIAL.get("instance/id") == self.ds.get_instance_id()

        shostname = GCE_META_PARTIAL.get("instance/hostname").split(".")[0]
        assert shostname == self.ds.get_hostname().hostname

    @responses.activate
    def test_userdata_no_encoding(self):
        """check that user-data is read."""
        self._set_mock_metadata(GCE_USER_DATA_TEXT)
        self.ds.get_data()
        assert (
            GCE_USER_DATA_TEXT["instance/attributes"]["user-data"].encode()
            == self.ds.get_userdata_raw()
        )

    @responses.activate
    def test_metadata_encoding(self):
        """user-data is base64 encoded if user-data-encoding is 'base64'."""
        self._set_mock_metadata(GCE_META_ENCODING)
        self.ds.get_data()

        instance_data = GCE_META_ENCODING.get("instance/attributes")
        decoded = b64decode(instance_data.get("user-data"))
        assert decoded == self.ds.get_userdata_raw()

    @responses.activate
    def test_missing_required_keys_return_false(self):
        for required_key in [
            "instance/id",
            "instance/zone",
            "instance/hostname",
        ]:
            meta = GCE_META_PARTIAL.copy()
            del meta[required_key]
            self._set_mock_metadata(meta)
            assert False is self.ds.get_data()
            responses.reset()

    @responses.activate
    def test_no_ssh_keys_metadata(self):
        self._set_mock_metadata()
        self.ds.get_data()
        assert [] == self.ds.get_public_ssh_keys()

    @responses.activate
    def test_cloudinit_ssh_keys(self):
        valid_key = "ssh-rsa VALID {0}"
        invalid_key = "ssh-rsa INVALID {0}"
        project_attributes = {
            "sshKeys": "\n".join(
                [
                    "cloudinit:{0}".format(valid_key.format(0)),
                    "user:{0}".format(invalid_key.format(0)),
                ]
            ),
            "ssh-keys": "\n".join(
                [
                    "cloudinit:{0}".format(valid_key.format(1)),
                    "user:{0}".format(invalid_key.format(1)),
                ]
            ),
        }
        instance_attributes = {
            "ssh-keys": "\n".join(
                [
                    "cloudinit:{0}".format(valid_key.format(2)),
                    "user:{0}".format(invalid_key.format(2)),
                ]
            ),
            "block-project-ssh-keys": "False",
        }

        meta = GCE_META.copy()
        meta["project/attributes"] = project_attributes
        meta["instance/attributes"] = instance_attributes

        self._set_mock_metadata(meta)
        self.ds.get_data()

        expected = [valid_key.format(key) for key in range(3)]
        assert set(expected) == set(self.ds.get_public_ssh_keys())

    @responses.activate
    @mock.patch(M_PATH + "ug_util")
    def test_default_user_ssh_keys(self, mock_ug_util, paths):
        mock_ug_util.normalize_users_groups.return_value = None, None
        mock_ug_util.extract_default.return_value = "ubuntu", None
        ubuntu_ds = DataSourceGCE.DataSourceGCE(
            settings.CFG_BUILTIN,
            self._make_distro("ubuntu"),
            paths,
        )

        valid_key = "ssh-rsa VALID {0}"
        invalid_key = "ssh-rsa INVALID {0}"
        project_attributes = {
            "sshKeys": "\n".join(
                [
                    "ubuntu:{0}".format(valid_key.format(0)),
                    "user:{0}".format(invalid_key.format(0)),
                ]
            ),
            "ssh-keys": "\n".join(
                [
                    "ubuntu:{0}".format(valid_key.format(1)),
                    "user:{0}".format(invalid_key.format(1)),
                ]
            ),
        }
        instance_attributes = {
            "ssh-keys": "\n".join(
                [
                    "ubuntu:{0}".format(valid_key.format(2)),
                    "user:{0}".format(invalid_key.format(2)),
                ]
            ),
            "block-project-ssh-keys": "False",
        }

        meta = GCE_META.copy()
        meta["project/attributes"] = project_attributes
        meta["instance/attributes"] = instance_attributes

        self._set_mock_metadata(meta)
        ubuntu_ds.get_data()

        expected = [valid_key.format(key) for key in range(3)]
        assert set(expected) == set(ubuntu_ds.get_public_ssh_keys())

    @responses.activate
    def test_instance_ssh_keys_override(self):
        valid_key = "ssh-rsa VALID {0}"
        invalid_key = "ssh-rsa INVALID {0}"
        project_attributes = {
            "sshKeys": "cloudinit:{0}".format(invalid_key.format(0)),
            "ssh-keys": "cloudinit:{0}".format(invalid_key.format(1)),
        }
        instance_attributes = {
            "sshKeys": "cloudinit:{0}".format(valid_key.format(0)),
            "ssh-keys": "cloudinit:{0}".format(valid_key.format(1)),
            "block-project-ssh-keys": "False",
        }

        meta = GCE_META.copy()
        meta["project/attributes"] = project_attributes
        meta["instance/attributes"] = instance_attributes

        self._set_mock_metadata(meta)
        self.ds.get_data()

        expected = [valid_key.format(key) for key in range(2)]
        assert set(expected) == set(self.ds.get_public_ssh_keys())

    @responses.activate
    def test_block_project_ssh_keys_override(self):
        valid_key = "ssh-rsa VALID {0}"
        invalid_key = "ssh-rsa INVALID {0}"
        project_attributes = {
            "sshKeys": "cloudinit:{0}".format(invalid_key.format(0)),
            "ssh-keys": "cloudinit:{0}".format(invalid_key.format(1)),
        }
        instance_attributes = {
            "ssh-keys": "cloudinit:{0}".format(valid_key.format(0)),
            "block-project-ssh-keys": "True",
        }

        meta = GCE_META.copy()
        meta["project/attributes"] = project_attributes
        meta["instance/attributes"] = instance_attributes

        self._set_mock_metadata(meta)
        self.ds.get_data()

        expected = [valid_key.format(0)]
        assert set(expected) == set(self.ds.get_public_ssh_keys())

    @responses.activate
    def test_only_last_part_of_zone_used_for_availability_zone(self):
        self._set_mock_metadata()
        assert True is self.ds.get_data()
        assert "bar" == self.ds.availability_zone

    @mock.patch("cloudinit.sources.DataSourceGCE.GoogleMetadataFetcher")
    def test_get_data_returns_false_if_not_on_gce(self, m_fetcher):
        self.m_platform_reports_gce.return_value = False
        ret = self.ds.get_data()
        assert False is ret
        m_fetcher.assert_not_called()

    def test_has_expired(self):
        def _get_timestamp(days):
            format_str = "%Y-%m-%dT%H:%M:%S%z"
            today = datetime.datetime.now(datetime.timezone.utc)
            timestamp = today + datetime.timedelta(days=days)
            return timestamp.strftime(format_str)

        past = _get_timestamp(-1)
        future = _get_timestamp(1)
        ssh_keys = {
            None: False,
            "": False,
            "Invalid": False,
            "user:ssh-rsa key user@domain.com": False,
            'user:ssh-rsa key google {"expireOn":"%s"}' % past: False,
            "user:ssh-rsa key google-ssh": False,
            "user:ssh-rsa key google-ssh {invalid:json}": False,
            'user:ssh-rsa key google-ssh {"userName":"user"}': False,
            'user:ssh-rsa key google-ssh {"expireOn":"invalid"}': False,
            'user:xyz key google-ssh {"expireOn":"%s"}' % future: False,
            'user:xyz key google-ssh {"expireOn":"%s"}' % past: True,
        }

        for key, expired in ssh_keys.items():
            assert DataSourceGCE._has_expired(key) == expired

    def test_parse_public_keys_non_ascii(self):
        public_key_data = [
            "cloudinit:rsa ssh-ke%s invalid" % chr(165),
            "use%sname:rsa ssh-key" % chr(174),
            "cloudinit:test 1",
            "default:test 2",
            "user:test 3",
        ]
        expected = ["test 1", "test 2"]
        found = DataSourceGCE._parse_public_keys(
            public_key_data, default_user="default"
        )
        assert sorted(found) == sorted(expected)

    @mock.patch("cloudinit.url_helper.readurl")
    def test_publish_host_keys(self, m_readurl):
        hostkeys = [("ssh-rsa", "asdfasdf"), ("ssh-ed25519", "qwerqwer")]
        readurl_expected_calls = [
            mock.call(
                check_status=False,
                data=b"asdfasdf",
                headers=HEADERS,
                request_method="PUT",
                url="%s%s" % (GUEST_ATTRIBUTES_URL, "ssh-rsa"),
            ),
            mock.call(
                check_status=False,
                data=b"qwerqwer",
                headers=HEADERS,
                request_method="PUT",
                url="%s%s" % (GUEST_ATTRIBUTES_URL, "ssh-ed25519"),
            ),
        ]
        self.ds.publish_host_keys(hostkeys)
        m_readurl.assert_has_calls(readurl_expected_calls, any_order=True)

    @responses.activate
    @mock.patch(
        M_PATH + "EphemeralDHCPv4",
        autospec=True,
    )
    @mock.patch(M_PATH + "net.find_candidate_nics", return_value=["ens4"])
    def test_local_datasource_uses_ephemeral_dhcp(
        self, _m_find_candidate_nics, m_dhcp, tmp_path
    ):
        self._set_mock_metadata()
        distro = mock.MagicMock()
        distro.get_tmp_exec_path = str(tmp_path)
        ds = DataSourceGCE.DataSourceGCELocal(
            sys_cfg={}, distro=distro, paths=None
        )
        ds._get_data()
        assert m_dhcp.call_count == 1

    @responses.activate
    @mock.patch(M_PATH + "read_md")
    @mock.patch(
        M_PATH + "EphemeralDHCPv4",
        autospec=True,
    )
    @mock.patch(M_PATH + "net.find_candidate_nics")
    def test_local_datasource_tries_on_multi_nic(
        self, m_find_candidate_nics, m_dhcp, m_read_md, caplog, tmp_path
    ):
        self._set_mock_metadata()
        distro = mock.MagicMock()
        distro.get_tmp_exec_path = str(tmp_path)
        ds = DataSourceGCE.DataSourceGCELocal(
            sys_cfg={}, distro=distro, paths=None
        )
        m_find_candidate_nics.return_value = [
            "ens0p4",
            "ens0p5",
            "ens0p6",
            "ens4",
        ]
        m_dhcp.return_value.__enter__.side_effect = (
            None,
            NoDHCPLeaseError,
            None,
            None,
        )
        m_read_md.side_effect = (
            {"success": False},
            Exception("whoopsie, not this one"),
            {
                "success": True,
                "platform_reports_gce": True,
                "reason": "reason",
                "user-data": "ud",
                "meta-data": "md",
            },
        )
        assert ds._get_data() is True
        assert m_dhcp.call_args_list == [
            mock.call(distro, iface=DataSourceGCE.DEFAULT_PRIMARY_INTERFACE),
            mock.call(distro, iface="ens0p4"),
            mock.call(distro, iface="ens0p5"),
            mock.call(distro, iface="ens0p6"),
        ]
        assert ds.distro.fallback_interface == "ens0p6"
        assert ds.metadata == "md"
        assert ds.userdata_raw == "ud"

        expected_logs = (
            "Looking for the primary NIC in:"
            " ['ens4', 'ens0p4', 'ens0p5', 'ens0p6']",
            "Error fetching IMD with candidate NIC ens0p5:"
            " whoopsie, not this one",
        )
        for msg in expected_logs:
            assert msg in caplog.text

    @responses.activate
    @mock.patch(
        M_PATH + "EphemeralDHCPv4",
        autospec=True,
    )
    def test_datasource_doesnt_use_ephemeral_dhcp(self, m_dhcp):
        self._set_mock_metadata()
        ds = DataSourceGCE.DataSourceGCE(sys_cfg={}, distro=None, paths=None)
        ds._get_data()
        assert m_dhcp.call_count == 0

    @responses.activate
    @mock.patch(
        M_PATH + "EphemeralDHCPv4",
        autospec=True,
    )
    @mock.patch(M_PATH + "net.find_candidate_nics")
    def test_datasource_on_dhcp_lease_failure(
        self, m_find_candidate_nics, m_dhcp, caplog, tmp_path
    ):
        self._set_mock_metadata()
        distro = mock.MagicMock()
        distro.get_tmp_exec_path = str(tmp_path)
        ds = DataSourceGCE.DataSourceGCELocal(
            sys_cfg={}, distro=distro, paths=None
        )
        m_find_candidate_nics.return_value = [
            "ens0p4",
            "ens0p5",
        ]
        m_dhcp.return_value.__enter__.side_effect = (
            NoDHCPLeaseError,
            NoDHCPLeaseError,
        )
        assert ds._get_data() is False
        assert m_dhcp.call_args_list == [
            mock.call(distro, iface="ens0p4"),
            mock.call(distro, iface="ens0p5"),
        ]

        expected_logs = (
            "Looking for the primary NIC in: ['ens0p4', 'ens0p5']",
            "Unable to obtain a DHCP lease for ens0p4",
            "Unable to obtain a DHCP lease for ens0p5",
        )
        for msg in expected_logs:
            assert msg in caplog.text

    @responses.activate
    def test_instance_and_project_data_decoded(self):
        """instance_data and project_data is decoded and can be queried."""
        md = {
            "instance/id": "123",
            "instance/zone": "foo/bar",
            "instance/hostname": "server.project-foo.local",
            "instance/attributes": {"ikey": "ivalue"},
            "project/attributes": {"pkey": "pvalue"},
        }

        self._set_mock_metadata(md)
        assert self.ds.get_data() is True

        assert "instance-data" in self.ds.metadata
        assert isinstance(self.ds.metadata["instance-data"], dict)
        assert "ikey" in self.ds.metadata["instance-data"]
        assert "ivalue" == self.ds.metadata["instance-data"]["ikey"]

        assert "project-data" in self.ds.metadata
        assert isinstance(self.ds.metadata["project-data"], dict)
        assert "pkey" in self.ds.metadata["project-data"]
        assert "pvalue" == self.ds.metadata["project-data"]["pkey"]



# This file is part of cloud-init. See LICENSE file for license information.
"""
This file is for testing the feature flag functionality itself,
NOT for testing any individual feature flag
"""
from unittest import mock

from cloudinit import features


class TestGetFeatures:
    def test_feature_without_override(self):
        # Since features are intended to be overridden downstream, mock them
        # all here so that downstream changes to features do not require
        # changes to this test.
        with mock.patch.multiple(
            "cloudinit.features",
            ERROR_ON_USER_DATA_FAILURE=True,
            ALLOW_EC2_MIRRORS_ON_NON_AWS_INSTANCE_TYPES=True,
            EXPIRE_APPLIES_TO_HASHED_USERS=False,
            NETPLAN_CONFIG_ROOT_READ_ONLY=True,
            DEPRECATION_INFO_BOUNDARY="devel",
            NOCLOUD_SEED_URL_APPEND_FORWARD_SLASH=False,
            APT_DEB822_SOURCE_LIST_FILE=True,
            MANUAL_NETWORK_WAIT=False,
            STRIP_INVALID_MTU=False,
        ):
            assert {
                "ERROR_ON_USER_DATA_FAILURE": True,
                "ALLOW_EC2_MIRRORS_ON_NON_AWS_INSTANCE_TYPES": True,
                "EXPIRE_APPLIES_TO_HASHED_USERS": False,
                "NETPLAN_CONFIG_ROOT_READ_ONLY": True,
                "NOCLOUD_SEED_URL_APPEND_FORWARD_SLASH": False,
                "APT_DEB822_SOURCE_LIST_FILE": True,
                "DEPRECATION_INFO_BOUNDARY": "devel",
                "MANUAL_NETWORK_WAIT": False,
                "STRIP_INVALID_MTU": False,
            } == features.get_features()



from unittest import mock

from cloudinit.net.netops import iproute2
from cloudinit.subp import SubpResult


class TestOps:
    @mock.patch.object(iproute2.subp, "subp")
    def test_link_up(self, m_subp):
        iproute2.Iproute2.link_up("eth0")
        iproute2.Iproute2.link_up("eth0", "inet6")
        assert m_subp.call_args_list == [
            mock.call(["ip", "link", "set", "dev", "eth0", "up"]),
            mock.call(
                ["ip", "-family", "inet6", "link", "set", "dev", "eth0", "up"]
            ),
        ]

    @mock.patch.object(iproute2.subp, "subp")
    def test_link_down(self, m_subp):
        iproute2.Iproute2.link_down("enp24s0")
        iproute2.Iproute2.link_down("eno1", "inet6")
        assert m_subp.call_args_list == [
            mock.call(["ip", "link", "set", "dev", "enp24s0", "down"]),
            mock.call(
                [
                    "ip",
                    "-family",
                    "inet6",
                    "link",
                    "set",
                    "dev",
                    "eno1",
                    "down",
                ]
            ),
        ]

    @mock.patch.object(iproute2.subp, "subp")
    def test_link_rename(self, m_subp):
        iproute2.Iproute2.link_rename("ens1", "ego1")
        assert m_subp.call_args_list == [
            mock.call(["ip", "link", "set", "ens1", "name", "ego1"])
        ]

    @mock.patch.object(iproute2.subp, "subp")
    def test_add_route(self, m_subp):
        iproute2.Iproute2.add_route("wlan0", "102.42.42.0/24")
        iproute2.Iproute2.add_route(
            "ens2",
            "102.42.0.0/16",
            gateway="192.168.2.254",
            source_address="192.168.2.1",
        )
        assert m_subp.call_args_list == [
            mock.call(
                [
                    "ip",
                    "-4",
                    "route",
                    "replace",
                    "102.42.42.0/24",
                    "dev",
                    "wlan0",
                ]
            ),
            mock.call(
                [
                    "ip",
                    "-4",
                    "route",
                    "replace",
                    "102.42.0.0/16",
                    "via",
                    "192.168.2.254",
                    "dev",
                    "ens2",
                    "src",
                    "192.168.2.1",
                ]
            ),
        ]

    @mock.patch.object(iproute2.subp, "subp")
    def test_del_route(self, m_subp):
        iproute2.Iproute2.del_route("wlan0", "102.42.42.0/24")
        iproute2.Iproute2.del_route(
            "ens2",
            "102.42.0.0/16",
            gateway="192.168.2.254",
            source_address="192.168.2.1",
        )
        assert m_subp.call_args_list == [
            mock.call(
                ["ip", "-4", "route", "del", "102.42.42.0/24", "dev", "wlan0"]
            ),
            mock.call(
                [
                    "ip",
                    "-4",
                    "route",
                    "del",
                    "102.42.0.0/16",
                    "via",
                    "192.168.2.254",
                    "dev",
                    "ens2",
                    "src",
                    "192.168.2.1",
                ]
            ),
        ]

    @mock.patch.object(iproute2.subp, "subp")
    def test_append_route(self, m_subp):
        iproute2.Iproute2.append_route("wlan0", "102.42.42.0/24", "10.0.4.254")
        assert m_subp.call_args_list == [
            mock.call(
                [
                    "ip",
                    "-4",
                    "route",
                    "append",
                    "102.42.42.0/24",
                    "via",
                    "10.0.4.254",
                    "dev",
                    "wlan0",
                ]
            )
        ]

    @mock.patch.object(iproute2.subp, "subp")
    def test_add_addr(self, m_subp):
        iproute2.Iproute2.add_addr("wlan0", "10.0.17.0", "10.0.17.255")
        assert m_subp.call_args_list == [
            mock.call(
                [
                    "ip",
                    "-family",
                    "inet",
                    "addr",
                    "add",
                    "10.0.17.0",
                    "broadcast",
                    "10.0.17.255",
                    "dev",
                    "wlan0",
                ],
            ),
        ]

    @mock.patch.object(iproute2.subp, "subp")
    def test_del_addr(self, m_subp):
        iproute2.Iproute2.del_addr("eth0", "10.0.8.3")
        assert m_subp.call_args_list == [
            mock.call(
                [
                    "ip",
                    "-family",
                    "inet",
                    "addr",
                    "del",
                    "10.0.8.3",
                    "dev",
                    "eth0",
                ],
            ),
        ]

    @mock.patch.object(iproute2.subp, "subp")
    def test_flush_addr(self, m_subp):
        iproute2.Iproute2.flush_addr("eth0")
        assert m_subp.call_args_list == [
            mock.call(
                ["ip", "addr", "flush", "dev", "eth0"],
            ),
        ]

    @mock.patch.object(
        iproute2.subp,
        "subp",
        return_value=SubpResult(
            "default via 192.168.0.1 dev enp2s0 proto dhcp src 192.168.0.104"
            " metric 100",
            "",
        ),
    )
    def test_add_default_route(self, m_subp):
        assert iproute2.Iproute2.get_default_route() == (
            "default via 192.168.0.1 dev enp2s0 proto dhcp src"
            " 192.168.0.104 metric 100"
        )
        assert m_subp.call_args_list == [
            mock.call(
                [
                    "ip",
                    "route",
                    "show",
                    "0.0.0.0/0",
                ],
            ),
        ]



import pytest

from tests.integration_tests.instances import IntegrationInstance
from tests.integration_tests.integration_settings import PLATFORM
from tests.integration_tests.releases import CURRENT_RELEASE, JAMMY

USER_DATA = """\
#cloud-config
runcmd:
 - echo "hi" >> /var/tmp/hi
snap:
 commands:
   00: echo heya
"""


@pytest.mark.user_data(USER_DATA)
def test_frequency_override(client: IntegrationInstance):
    # Some pre-checks
    assert (
        "running config-scripts_user with frequency once-per-instance"
        in client.read_from_file("/var/log/cloud-init.log")
    )
    assert client.read_from_file("/var/tmp/hi").strip().count("hi") == 1
    # This workaround is not needed for OCI, so just skip it
    if CURRENT_RELEASE.os == "ubuntu" and PLATFORM != "oci":
        if CURRENT_RELEASE > JAMMY:
            # Newer series will not block on snapd.seeded.service
            assert "snapd.seeded.service" not in client.execute(
                "systemctl show -P After cloud-config.service"
            )
            assert client.execute(
                "test -f /var/lib/cloud/sem/snap_seeded.once"
            ).ok
        else:
            # Jammy series and earlier will block on snapd.seeded.service
            assert "snapd.seeded.service" in client.execute(
                "systemctl show -P After cloud-config.service"
            )
            # Jammy will not create a snap-seeded semaphore.
            assert not client.execute(
                "test -f /var/lib/cloud/sem/snap_seeded.once"
            ).ok

    # Change frequency of scripts_user to always
    config = client.read_from_file("/etc/cloud/cloud.cfg")
    new_config = config.replace("- scripts_user", "- [ scripts_user, always ]")
    client.write_to_file("/etc/cloud/cloud.cfg", new_config)

    client.restart()

    # Ensure the script was run again
    assert (
        "running config-scripts_user with frequency always"
        in client.read_from_file("/var/log/cloud-init.log")
    )
    assert client.read_from_file("/var/tmp/hi").strip().count("hi") == 2



# Copyright (C) 2013 Hewlett-Packard Development Company, L.P.
#
# Author: Juerg Haefliger <juerg.haefliger@hp.com>
#
# This file is part of cloud-init. See LICENSE file for license information.
import logging
from io import BytesIO
from pathlib import Path

import pytest
from configobj import ConfigObj

from cloudinit import util
from cloudinit.config import cc_locale
from cloudinit.config.schema import (
    SchemaValidationError,
    get_schema,
    validate_cloudconfig_schema,
)
from tests.unittests.helpers import does_not_raise, mock, skipUnlessJsonSchema
from tests.unittests.util import get_cloud

LOG = logging.getLogger(__name__)


@pytest.mark.usefixtures("fake_filesystem")
class TestLocale:

    def test_set_locale_arch(self):
        locale = "en_GB.UTF-8"
        locale_configfile = "/etc/invalid-locale-path"
        cfg = {
            "locale": locale,
            "locale_configfile": locale_configfile,
        }
        cc = get_cloud("arch")

        with mock.patch("cloudinit.distros.arch.subp.subp") as m_subp:
            with mock.patch("cloudinit.distros.arch.LOG.warning") as m_LOG:
                cc_locale.handle("cc_locale", cfg, cc, [])
                m_LOG.assert_called_with(
                    "Invalid locale_configfile %s, "
                    "only supported value is "
                    "/etc/locale.conf",
                    locale_configfile,
                )

                contents = util.load_text_file(cc.distro.locale_gen_fn)
                assert "%s UTF-8" % locale in contents
                m_subp.assert_called_with(
                    ["localectl", "set-locale", locale], capture=False
                )

    @pytest.mark.parametrize(
        "distro_name,uses_systemd, cfg,expected_locale",
        (
            pytest.param(
                "sles",
                True,
                {},
                {"LANG": "en_US.UTF-8"},
                id="sles_locale_defaults_systemd",
            ),
            pytest.param(
                "sles",
                False,
                {},
                {"RC_LANG": "en_US.UTF-8"},
                id="sles_locale_defaults_not_systemd",
            ),
            pytest.param(
                "sles",
                True,
                {"locale": "My.Locale"},
                {"LANG": "My.Locale"},
                id="custom_locale",
            ),
            pytest.param(
                "rhel",
                True,
                {},
                {"LANG": "en_US.UTF-8"},
                id="test_rhel_default_locale",
            ),
        ),
    )
    def test_set_locale_sles(
        self, distro_name, uses_systemd, cfg, expected_locale, mocker
    ):
        cc = get_cloud(distro_name)
        mocker.patch.object(
            cc.distro, "uses_systemd", return_value=uses_systemd
        )
        cc_locale.handle("cc_locale", cfg, cc, [])
        if uses_systemd:
            locale_conf = cc.distro.systemd_locale_conf_fn
        else:
            locale_conf = cc.distro.locale_conf_fn
        contents = util.load_binary_file(locale_conf)
        n_cfg = ConfigObj(BytesIO(contents))
        assert expected_locale == dict(n_cfg)

    def test_locale_update_config_if_different_than_default(self, tmpdir):
        """Test cc_locale writes updates conf if different than default"""
        locale_conf = tmpdir.join("etc/default/locale")
        Path(locale_conf).parent.mkdir(parents=True)
        locale_conf.write('LANG="en_US.UTF-8"\n')
        cfg = {"locale": "C.UTF-8"}
        cc = get_cloud("ubuntu")
        with mock.patch("cloudinit.distros.debian.subp.subp") as m_subp:
            with mock.patch(
                "cloudinit.distros.debian.LOCALE_CONF_FN", locale_conf.strpath
            ):
                with mock.patch(
                    "cloudinit.distros.debian.subp.which",
                    return_value="/usr/sbin/update-locale",
                ) as m_which:
                    cc_locale.handle("cc_locale", cfg, cc, [])
        m_subp.assert_called_with(
            [
                "update-locale",
                "--locale-file=%s" % locale_conf.strpath,
                "LANG=C.UTF-8",
            ],
            capture=False,
        )
        m_which.assert_called_once_with("update-locale")


class TestLocaleSchema:
    @pytest.mark.parametrize(
        "config, expectation",
        (
            # Valid schemas tested via meta['examples'] in test_schema.py
            # Invalid schemas
            (
                {"locale": 1},
                pytest.raises(
                    SchemaValidationError,
                    match="locale: 1 is not of type 'string'",
                ),
            ),
            (
                {"locale_configfile": 1},
                pytest.raises(
                    SchemaValidationError,
                    match="locale_configfile: 1 is not of type 'string'",
                ),
            ),
            ({"locale": False}, does_not_raise()),
        ),
    )
    @skipUnlessJsonSchema()
    def test_schema_validation(self, config, expectation):
        schema = get_schema()
        with expectation:
            validate_cloudconfig_schema(config, schema, strict=True)



"""NoCloud datasource integration tests."""

from textwrap import dedent

import pytest
from pycloudlib.lxd.instance import LXDInstance

from cloudinit import lifecycle
from cloudinit.subp import subp
from cloudinit.util import is_true
from tests.integration_tests.instances import IntegrationInstance
from tests.integration_tests.integration_settings import PLATFORM
from tests.integration_tests.releases import CURRENT_RELEASE, FOCAL
from tests.integration_tests.util import (
    get_feature_flag_value,
    network_wait_logged,
    override_kernel_command_line,
    verify_clean_boot,
    verify_clean_log,
)

VENDOR_DATA = """\
#cloud-config
bootcmd:
  - touch /var/tmp/seeded_vendordata_test_file
"""

# The fallback network config doesn't work on LXD, leading to
# systemd-network-wait-online timing out and/or cloud-init raising
# an error about wait-online.
# This gives us a NoCloud default to work around these issues
NETWORK_CONFIG = """\
network:
  version: 2
  ethernets:
    eth0:
      dhcp4: true
"""


LXD_METADATA_NOCLOUD_SEED = """\
  /var/lib/cloud/seed/nocloud-net/meta-data:
    when:
    - create
    - copy
    create_only: false
    template: emptycfg.tpl
    properties:
      default: |
        #cloud-config
        {}
  /var/lib/cloud/seed/nocloud-net/user-data:
    when:
    - create
    - copy
    create_only: false
    template: emptycfg.tpl
    properties:
      default: |
        #cloud-config
        {}
  /var/lib/cloud/seed/nocloud-net/network-config:
    when:
    - create
    - copy
    create_only: false
    template: netcfg.tpl
"""


def setup_nocloud(instance: LXDInstance):
    # On Jammy and above, LXD no longer uses NoCloud, so we need to set
    # it up manually
    lxd_image_metadata = subp(
        ["lxc", "config", "metadata", "show", instance.name]
    )
    if "/var/lib/cloud/seed/nocloud-net" in lxd_image_metadata.stdout:
        return
    subp(
        ["lxc", "config", "template", "create", instance.name, "emptycfg.tpl"],
    )
    subp(
        ["lxc", "config", "template", "edit", instance.name, "emptycfg.tpl"],
        data="#cloud-config\n{}\n",
    )
    subp(
        ["lxc", "config", "template", "create", instance.name, "netcfg.tpl"],
    )
    subp(
        ["lxc", "config", "template", "edit", instance.name, "netcfg.tpl"],
        data=NETWORK_CONFIG,
    )
    subp(
        ["lxc", "config", "metadata", "edit", instance.name],
        data=f"{lxd_image_metadata.stdout}{LXD_METADATA_NOCLOUD_SEED}",
    )


@pytest.mark.lxd_setup.with_args(setup_nocloud)
@pytest.mark.lxd_use_exec
@pytest.mark.skipif(
    PLATFORM != "lxd_container",
    reason="Requires NoCloud with custom setup",
)
def test_nocloud_seedfrom_vendordata(client: IntegrationInstance):
    """Integration test for #570.

    Test that we can add optional vendor-data to the seedfrom file in a
    NoCloud environment
    """
    seed_dir = "/var/tmp/test_seed_dir"
    result = client.execute(
        "mkdir {seed_dir} && "
        "touch {seed_dir}/user-data && "
        "touch {seed_dir}/meta-data && "
        "echo '{net}' > {seed_dir}/network-config && "
        "echo 'seedfrom: {seed_dir}/' > "
        "/var/lib/cloud/seed/nocloud-net/meta-data".format(
            seed_dir=seed_dir, net=NETWORK_CONFIG
        )
    )
    assert result.return_code == 0

    client.write_to_file(
        "{}/vendor-data".format(seed_dir),
        VENDOR_DATA,
    )
    client.execute("cloud-init clean --logs")
    client.restart()
    assert client.execute("cloud-init status").ok
    assert "seeded_vendordata_test_file" in client.execute("ls /var/tmp")
    assert network_wait_logged(
        client.execute("cat /var/log/cloud-init.log")
    ) == is_true(get_feature_flag_value(client, "MANUAL_NETWORK_WAIT"))


SMBIOS_USERDATA = """\
#cloud-config
runcmd:
  - touch /var/tmp/smbios_test_file
"""
SMBIOS_SEED_DIR = "/smbios_seed"


def setup_nocloud_local_serial(instance: LXDInstance):
    subp(
        [
            "lxc",
            "config",
            "set",
            instance.name,
            "raw.qemu=-smbios "
            f"type=1,serial=ds=nocloud;s=file://{SMBIOS_SEED_DIR};h=myhost",
        ]
    )


def setup_nocloud_network_serial(instance: LXDInstance):
    subp(
        [
            "lxc",
            "config",
            "set",
            instance.name,
            "raw.qemu=-smbios "
            "type=1,serial=ds=nocloud-net;s=http://0.0.0.0/;h=myhost",
        ]
    )


@pytest.mark.lxd_use_exec
@pytest.mark.skipif(
    PLATFORM != "lxd_vm",
    reason="Requires NoCloud with raw QEMU serial setup",
)
class TestSmbios:
    @pytest.mark.lxd_setup.with_args(setup_nocloud_local_serial)
    def test_smbios_seed_local(self, client: IntegrationInstance):
        """Check that smbios seeds that use local disk work"""
        assert client.execute(f"mkdir -p {SMBIOS_SEED_DIR}").ok
        client.write_to_file(f"{SMBIOS_SEED_DIR}/user-data", SMBIOS_USERDATA)
        client.write_to_file(f"{SMBIOS_SEED_DIR}/meta-data", "")
        client.write_to_file(f"{SMBIOS_SEED_DIR}/vendor-data", "")
        assert client.execute("cloud-init clean --logs").ok
        client.restart()
        assert client.execute("test -f /var/tmp/smbios_test_file").ok

    @pytest.mark.lxd_setup.with_args(setup_nocloud_network_serial)
    def test_smbios_seed_network(self, client: IntegrationInstance):
        """Check that smbios seeds that use network (http/https) work"""
        service_file = "/lib/systemd/system/local-server.service"
        client.write_to_file(
            service_file,
            dedent(
                """\
                [Unit]
                Description=Serve a local webserver
                Before=cloud-init-network.service
                Wants=cloud-init-local.service
                DefaultDependencies=no
                After=systemd-networkd-wait-online.service
                After=networking.service


                [Install]
                WantedBy=cloud-init.target

                [Service]
                """
                f"WorkingDirectory={SMBIOS_SEED_DIR}"
                """
                ExecStart=/usr/bin/env python3 -m http.server --bind 0.0.0.0 80
                """
            ),
        )
        assert client.execute(
            "chmod 644 /lib/systemd/system/local-server.service"
        ).ok
        assert client.execute("systemctl enable local-server.service").ok
        client.write_to_file(
            "/etc/cloud/cloud.cfg.d/91_do_not_use_lxd.cfg",
            "datasource_list: [ NoCloud, None ]\n",
        )
        assert client.execute(f"mkdir -p {SMBIOS_SEED_DIR}").ok
        client.write_to_file(f"{SMBIOS_SEED_DIR}/user-data", SMBIOS_USERDATA)
        client.write_to_file(f"{SMBIOS_SEED_DIR}/meta-data", "")
        client.write_to_file(f"{SMBIOS_SEED_DIR}/vendor-data", "")
        assert client.execute("cloud-init clean --logs").ok
        client.restart()
        assert client.execute("test -f /var/tmp/smbios_test_file").ok
        version_boundary = get_feature_flag_value(
            client, "DEPRECATION_INFO_BOUNDARY"
        )
        message = "The 'nocloud-net' datasource name is deprecated"
        # nocloud-net deprecated in version 24.1
        if lifecycle.should_log_deprecation("24.1", version_boundary):
            verify_clean_boot(client, require_deprecations=[message])
        else:
            client.execute(
                rf"grep \"INFO]: {message}\" /var/log/cloud-init.log"
            ).ok


@pytest.mark.skipif(PLATFORM != "lxd_vm", reason="Modifies grub config")
@pytest.mark.lxd_use_exec
class TestFTP:
    """Test nocloud's support for unencrypted FTP and FTP over TLS (ftps).

    These tests work by setting up a local ftp server on the test instance
    and then rebooting the instance clean (cloud-init clean --logs --reboot).

    Check for the existence (or non-existence) of specific log messages to
    verify functionality.
    """

    # should we really be surfacing this netplan stderr as a warning?
    # i.e. how does it affect the users?
    expected_warnings = [
        "Falling back to a hard restart of systemd-networkd.service"
    ]

    @staticmethod
    def _boot_with_cmdline(
        cmdline: str, client: IntegrationInstance, encrypted: bool = False
    ) -> None:
        """configure an ftp server to start prior to network timeframe
        optionally install certs and make the server support only FTP over TLS

        cmdline: a string containing the kernel command line set on reboot
        client: an instance to configure
        encrypted: a boolean which modifies the configured ftp server
        """

        # install the essential bits
        assert client.execute(
            "apt update && apt install -yq python3-pyftpdlib "
            "python3-openssl ca-certificates libnss3-tools"
        ).ok

        # How do you reliably run a ftp server for your instance to
        # read files from during early boot? In typical production
        # environments, the ftp server would be separate from the instance.
        #
        # For a reliable server that fits with the framework of running tests
        # on a single instance, it is easier to just install an ftp server
        # that runs on the second boot prior to the cloud-init unit which
        # reaches out to the ftp server. This achieves reaching out to an
        # ftp(s) server for testing - cloud-init just doesn't have to reach
        # very far to get what it needs.
        #
        # DO NOT use these concepts in a production.
        #
        # This configuration is neither secure nor production-grade - intended
        # only for testing purposes.
        client.write_to_file(
            "/server.py",
            dedent(
                """\
                #!/usr/bin/python3
                import logging

                from systemd.daemon import notify

                from pyftpdlib.authorizers import DummyAuthorizer
                from pyftpdlib.handlers import FTPHandler, TLS_FTPHandler
                from pyftpdlib.servers import FTPServer
                from pyftpdlib.filesystems import UnixFilesystem

                encrypted = """
                + str(encrypted)
                + """

                logging.basicConfig(level=logging.DEBUG)

                # yeah, it's not secure but that's not the point
                authorizer = DummyAuthorizer()

                # Define a read-only anonymous user
                authorizer.add_anonymous("/home/anonymous")

                # Instantiate FTP handler class
                if not encrypted:
                    handler = FTPHandler
                    logging.info("Running unencrypted ftp server")
                else:
                    handler = TLS_FTPHandler
                    handler.certfile = "/cert.pem"
                    handler.keyfile = "/key.pem"
                    logging.info("Running encrypted ftp server")

                handler.authorizer = authorizer
                handler.abstracted_fs = UnixFilesystem
                server = FTPServer(("localhost", 2121), handler)

                # tell systemd to proceed
                notify("READY=1")

                # start the ftp server
                server.serve_forever()
                """
            ),
        )
        assert client.execute("chmod +x /server.py").ok

        if encrypted:
            if CURRENT_RELEASE > FOCAL:
                assert client.execute("apt install -yq mkcert").ok
            else:

                # install golang
                assert client.execute("apt install -yq golang").ok

                # build mkcert from source
                #
                # we could check out a tag, but the project hasn't
                # been updated in 2 years
                #
                # instructions from https://github.com/FiloSottile/mkcert
                assert client.execute(
                    "git clone https://github.com/FiloSottile/mkcert && "
                    "cd mkcert && "
                    "export latest_ver=$(git describe --tags --abbrev=0) && "
                    'curl -JL "https://github.com/FiloSottile/mkcert/releases/'
                    "download/${latest_ver}/mkcert-"
                    '${latest_ver}-linux-amd64"'
                    " -o mkcert && "
                    "chmod 755 mkcert"
                ).ok

                # giddyup
                assert client.execute(
                    "ln -s $HOME/mkcert/mkcert /usr/local/bin/mkcert"
                ).ok

            # more palatable than openssl commands
            assert client.execute(
                "mkcert -install -cert-file /cert.pem -key-file /key.pem "
                "localhost 127.0.0.1 0.0.0.0 ::1"
            ).ok

        client.write_to_file(
            "/lib/systemd/system/local-ftp.service",
            dedent(
                """\
                [Unit]
                Description=TESTING USE ONLY ftp server
                Wants=cloud-init-local.service
                DefaultDependencies=no

                # we want the network up for network operations
                # and NoCloud operates in network timeframe
                After=systemd-networkd-wait-online.service
                After=networking.service
                Before=cloud-init-network.service
                Before=cloud-init.service

                [Service]
                Type=notify
                ExecStart=/server.py

                [Install]
                WantedBy=cloud-init.target
                """
            ),
        )
        assert client.execute(
            "chmod 644 /lib/systemd/system/local-ftp.service"
        ).ok
        assert client.execute("systemctl enable local-ftp.service").ok
        assert client.execute("mkdir /home/anonymous").ok

        client.write_to_file(
            "/user-data",
            dedent(
                """\
                #cloud-config

                hostname: ftp-bootstrapper
                """
            ),
        )
        client.write_to_file(
            "/meta-data",
            dedent(
                """\
                instance-id: ftp-instance
                """
            ),
        )
        client.write_to_file("/vendor-data", "")

        # set the kernel command line, reboot with it
        override_kernel_command_line(cmdline, client)

    def test_nocloud_ftp_unencrypted_server_succeeds(
        self, client: IntegrationInstance
    ):
        """check that ftp:// succeeds to unencrypted ftp server

        this mode allows administrators to choose unencrypted ftp,
        at their own risk
        """
        cmdline = "ds=nocloud;seedfrom=ftp://0.0.0.0:2121"
        self._boot_with_cmdline(cmdline, client)
        verify_clean_boot(client, ignore_warnings=self.expected_warnings)
        assert "ftp-bootstrapper" == client.execute("hostname").rstrip()
        verify_clean_log(client.execute("cat /var/log/cloud-init.log").stdout)

    def test_nocloud_ftps_unencrypted_server_fails(
        self, client: IntegrationInstance
    ):
        """check that ftps:// fails to unencrypted ftp server

        this mode allows administrators to enforce TLS encryption
        """
        cmdline = "ds=nocloud;seedfrom=ftps://localhost:2121"
        self._boot_with_cmdline(cmdline, client)
        verify_clean_boot(
            client,
            ignore_warnings=self.expected_warnings,
            require_warnings=[
                "Getting data from <class 'cloudinit.sources.DataSourc"
                "eNoCloud.DataSourceNoCloudNet'> failed",
                "Used fallback datasource",
                "Attempted to connect to an insecure ftp server but used"
                " a scheme of ftps://, which is not allowed. Use ftp:// "
                "to allow connecting to insecure ftp servers.",
            ],
            ignore_tracebacks=[
                'ftplib.error_perm: 500 Command "AUTH" not understood.',
                "UrlError: Attempted to connect to an insecure ftp server",
            ],
        )

    def test_nocloud_ftps_encrypted_server_succeeds(
        self, client: IntegrationInstance
    ):
        """check that ftps:// encrypted ftp server succeeds

        this mode allows administrators to enforce TLS encryption
        """
        cmdline = "ds=nocloud;seedfrom=ftps://localhost:2121"
        self._boot_with_cmdline(cmdline, client, encrypted=True)
        verify_clean_boot(client, ignore_warnings=self.expected_warnings)
        assert "ftp-bootstrapper" == client.execute("hostname").rstrip()
        verify_clean_log(client.execute("cat /var/log/cloud-init.log").stdout)

    def test_nocloud_ftp_encrypted_server_fails(
        self, client: IntegrationInstance
    ):
        """check that using ftp:// to encrypted ftp server fails"""
        cmdline = "ds=nocloud;seedfrom=ftp://0.0.0.0:2121"
        self._boot_with_cmdline(cmdline, client, encrypted=True)
        verify_clean_boot(client, ignore_warnings=self.expected_warnings)



# This file is part of cloud-init. See LICENSE file for license information.
import logging
import os
import stat
from typing import Any, Dict
from unittest.mock import patch

import pytest

from cloudinit import helpers, util
from cloudinit.config.cc_runcmd import handle
from cloudinit.config.schema import (
    SchemaValidationError,
    get_schema,
    validate_cloudconfig_schema,
)
from tests.unittests.helpers import SCHEMA_EMPTY_ERROR, skipUnlessJsonSchema
from tests.unittests.util import get_cloud

LOG = logging.getLogger(__name__)


@pytest.fixture
def cloud(tmp_path):
    paths = helpers.Paths({"scripts": str(tmp_path)})
    return get_cloud(paths=paths)


@pytest.mark.usefixtures("fake_filesystem")
class TestRuncmd:
    # self.paths = helpers.Paths({"scripts": self.new_root})

    def test_handler_skip_if_no_runcmd(self, caplog, cloud):
        """When the provided config doesn't contain runcmd, skip it."""
        cfg: Dict[str, Any] = {}
        handle("notimportant", cfg, cloud, [])
        assert (
            "Skipping module named notimportant, no 'runcmd' key"
            in caplog.text
        )

    @pytest.mark.allow_subp_for("/bin/sh")
    @patch("cloudinit.util.shellify")
    def test_runcmd_shellify_fails(self, cls, cloud):
        """When shellify fails throw exception"""
        cls.side_effect = TypeError("patched shellify")
        valid_config = {"runcmd": ["echo 42"]}
        with pytest.raises(TypeError, match="Failed to shellify"):
            handle("cc_runcmd", valid_config, cloud, [])

    def test_handler_invalid_command_set(self, cloud):
        """Commands which can't be converted to shell will raise errors."""
        invalid_config = {"runcmd": 1}
        with pytest.raises(
            TypeError,
            match="Failed to shellify 1 into file"
            " /var/lib/cloud/instances/iid-datasource-none/scripts/runcmd",
        ):
            handle("cc_runcmd", invalid_config, cloud, [])

    def test_handler_write_valid_runcmd_schema_to_file(self, cloud, tmp_path):
        """Valid runcmd schema is written to a runcmd shell script."""
        valid_config = {"runcmd": [["ls", "/"]]}
        handle("cc_runcmd", valid_config, cloud, [])
        runcmd_file = os.path.join(
            tmp_path,
            "var/lib/cloud/instances/iid-datasource-none/scripts/runcmd",
        )
        assert "#!/bin/sh\n'ls' '/'\n" == util.load_text_file(runcmd_file)
        file_stat = os.stat(runcmd_file)
        assert 0o700 == stat.S_IMODE(file_stat.st_mode)


@skipUnlessJsonSchema()
class TestRunCmdSchema:
    @pytest.mark.parametrize(
        "config, error_msg",
        (
            # Ensure duplicate commands are valid
            ({"runcmd": [["echo", "bye"], ["echo", "bye"]]}, None),
            ({"runcmd": ["echo bye", "echo bye"]}, None),
            # Invalid schemas
            ({"runcmd": 1}, "1 is not of type 'array'"),
            ({"runcmd": []}, rf"runcmd: \[\] {SCHEMA_EMPTY_ERROR}"),
            (
                {
                    "runcmd": [
                        "ls /",
                        20,
                        ["wcurl", "http://stuff/blah"],
                        {"a": "n"},
                    ]
                },
                "is not of type",
            ),
        ),
    )
    @skipUnlessJsonSchema()
    def test_schema_validation(self, config, error_msg):
        if error_msg is None:
            validate_cloudconfig_schema(config, get_schema(), strict=True)
        else:
            with pytest.raises(SchemaValidationError, match=error_msg):
                validate_cloudconfig_schema(config, get_schema(), strict=True)
